import uuid
from datetime import datetime
from sqlalchemy import String, Integer, Column, create_engine, ForeignKey, TIMESTAMP
from sqlalchemy.orm import sessionmaker, relationship, joinedload, subqueryload, Session as sqlalchemySession
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy_mptt import mptt_sessionmaker

import config as user_config

Engine = create_engine(
    user_config.DB_SQLALCHEMY_DATABASE_URL if hasattr(user_config, "DB_SQLALCHEMY_DATABASE_URL") else "sqlite:///db.sqlite3",
    **user_config.DB_ENGINE_KWARGS if hasattr(user_config, "DB_ENGINE_KWARGS") else {
        "connect_args": {
            'check_same_thread': False
        }
    }
)
SessionLocal = sessionmaker(bind=Engine, **user_config.DB_SESSION_MAKER_KWARGS if hasattr(user_config, "DB_SESSION_MAKER_KWARGS") else {
    "autoflush": False,
    "autocommit": False,
    "expire_on_commit": True
})

SessionMpttLocal = mptt_sessionmaker(SessionLocal)
# 创建基本映射类
BaseModel = declarative_base()


def session():
    """
    实例 sessionmaker
    :return:
    """
    _db = SessionLocal()
    try:
        yield _db
    finally:
        try:
            _db.close()
        except:
            pass


def mptt_session():
    """
    实例 sessionmaker
    :return:
    """
    mptt_db = SessionMpttLocal()
    try:
        yield mptt_db
    finally:
        mptt_db.close()


class Model(BaseModel):
    __abstract__ = True
    id = Column(Integer, primary_key=True, autoincrement=True)
    uuid = Column(String(32), unique=True, nullable=False, default=lambda: uuid.uuid4().hex, comment="UUID")
    deleted_at = Column(TIMESTAMP, index=True, nullable=True, comment="删除日期")
    created_at = Column(TIMESTAMP, index=True, nullable=True, default=datetime.now, comment="创建日期")
    updated_at = Column(TIMESTAMP, index=True, nullable=True, default=datetime.now, onupdate=datetime.now, comment="更新日期")

    def to_dict(self):
        """
        ORM转dict
        :return:
        """
        columns = [c.name for c in self.__table__.columns] + [name for name, obj in vars(self.__class__).items() if isinstance(obj, property)]
        return {key: getattr(self, key, None) for key in columns}



from datetime import datetime
from typing import Optional, List
from pydantic import BaseModel

from yao.schema import SchemaPrefixNames, SchemasPaginate


class SchemasFunctionAnnexeResponse(SchemaPrefixNames):
    """角色 返回"""
    md5: Optional[str] = None
    path: Optional[str] = None
    preview_path: Optional[str] = None
    size: Optional[int] = None
    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态

    class Config:
        orm_mode = True


class SchemasFunctionAnnexePaginateItem(SchemasPaginate):
    items: List[SchemasFunctionAnnexeResponse]


class SchemasFunctionAnnexeStoreUpdate(BaseModel):
    """提交"""
    prefix: Optional[str] = None
    filename: Optional[str] = None
    content_type: Optional[str] = None
    md5: Optional[str] = None
    path: Optional[str] = None
    size: Optional[int] = None
    width: Optional[int] = None
    height: Optional[int] = None
    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态


class SchemasParams(BaseModel):
    pass


class SchemasUpLoadFileResponse(BaseModel):
    state: Optional[str] = "SUCCESS"  # 上传状态，上传成功时必须返回"SUCCESS"
    url: Optional[str] = None  # 返回的地址
    title: Optional[str] = None  # 新文件名
    original: Optional[str] = None  # 原始文件名
    type: Optional[str] = None  # 文件类型
    size: Optional[str] = None  # 文件大小
    list: Optional[List[dict]] = None


class SchemasUpLoadContentFile(BaseModel):
    path: str  # 目录
    content: str  # 内容



import json
import asyncio
from datetime import datetime
from functools import wraps
from typing import Optional, List, Union

from fastapi import Depends, Security, APIRouter, Query, Request, HTTPException, status
from fastapi.security import SecurityScopes, OAuth2PasswordBearer
from fastapi.security.utils import get_authorization_scheme_param
from sqlalchemy.orm import Session

from config import OAUTH_TOKEN_URL, OAUTH_TOKEN_SCOPES, OAUTH_SECRET_KEY, OAUTH_ALGORITHM, OAUTH_LOGIN_SCOPES
from yao.db import session
from yao.helpers import token_payload
from yao.schema import ModelScreenParams, ModelScreenParamsForAll
from yao.function.user.crud import CrudFunctionUser
from yao.function.user.schema import SchemasFunctionUser, SchemasFunctionScopes
from yao.function.log.crud import CrudFunctionLog
from yao.function.log.schema import SchemasFunctionStoreUpdate as SchemasStoreUpdateLog


def model_post_screen_params(data: ModelScreenParams = None):
    """列表筛选参数"""
    order = []
    [order.extend(list(s.items())) for s in data.order]
    data.order = order
    where = [(w['key'], w['condition'], w['value']) for w in data.where if
             ('value' in w and (w['value'] or w['value'] is False or w['value'] == 0)) and (not "join" in w or "join" in w and not w['join'])]
    join = [(w['join'], [(w['key'], w['condition'], w['value'])], 'join') for w in data.where if
            ('value' in w and (w['value'] or w['value'] is False or w['value'] == 0)) and ("join" in w and w['join'])]
    data.where = where
    data.join = join
    return data


def model_post_screen_params_for_all(data: ModelScreenParamsForAll = None):
    """列表筛选参数"""
    order = []
    [order.extend(list(s.items())) for s in data.order]
    data.order = order
    where = [(w['key'], w['condition'], w['value']) for w in data.where if
             ('value' in w and (w['value'] or w['value'] is False or w['value'] == 0)) and (not "join" in w or "join" in w and not w['join'])]
    join = [(w['join'], [(w['key'], w['condition'], w['value'])], 'join') for w in data.where if
            ('value' in w and (w['value'] or w['value'] is False or w['value'] == 0)) and ("join" in w and w['join'])]
    data.where = where
    data.join = join
    return data


def model_screen_params(page: Optional[str] = None, limit: Optional[str] = None, where: Optional[str] = None, join: Optional[str] = None,
                        order: Optional[str] = None):
    """列表筛选参数"""
    data = ModelScreenParams(page=json.loads(page).get('value') if page else 1, limit=json.loads(limit).get('value') if limit else 25,
                             where=json.loads(where).get('value') if where else [], join=json.loads(join).get('value') if join else [],
                             order=json.loads(order).get('value') if order else [])

    return model_post_screen_params(data)


def model_screen_params_for_all(where: Optional[str] = None, join: Optional[str] = None,
                                order: Optional[str] = None):
    """列表筛选参数"""
    data = ModelScreenParamsForAll(where=json.loads(where).get('value') if where else [], join=json.loads(join).get('value') if join else [],
                                   order=json.loads(order).get('value') if order else [])

    return model_post_screen_params_for_all(data)


class OAuth2PasswordBearerOrForm(OAuth2PasswordBearer):
    async def __call__(self, request: Request) -> Optional[str]:
        authorization = request.headers.get("Authorization") or (await request.form()).get("Authorization")
        scheme, param = get_authorization_scheme_param(authorization)
        if not authorization or scheme.lower() != "bearer":
            if self.auto_error:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="Not authenticated",
                    headers={"WWW-Authenticate": "Bearer"},
                )
            else:
                return None
        return param


oauth2_scheme = OAuth2PasswordBearerOrForm(tokenUrl=OAUTH_TOKEN_URL, scopes=OAUTH_TOKEN_SCOPES)


async def current_user_security(security_scopes: SecurityScopes, token: str = Depends(oauth2_scheme)):
    """
    解析加密字段
    :param security_scopes:
    :param token:
    :return:
    """
    payload = token_payload(security_scopes, token, OAUTH_SECRET_KEY, OAUTH_ALGORITHM)
    """处理授权用户实时情况"""
    # Todo
    """处理授权用户实时情况"""
    return SchemasFunctionUser(**payload)


async def auth_user(auth: SchemasFunctionUser = Security(current_user_security, scopes=[OAUTH_LOGIN_SCOPES]),
                    session: Session = Depends(session)):
    """
    demo
    :param auth:
    :param session:
    :return:
    """
    user = CrudFunctionUser.init().first(session=session, pk=auth.user_id)
    return SchemasFunctionScopes(user=user, prefix=auth.prefix, scopes=auth.scopes, children_ids=user.children_ids)


def item_prefix(callback):
    """
    处理 item prefix
    :param callback:
    :return:
    """

    @wraps(callback)
    async def wrapper(*args, **kwargs):
        if "item" in kwargs and "auth" in kwargs:
            item = kwargs.get("item")
            auth = kwargs.get("auth")
            item.prefix = item.prefix or auth.prefix
            kwargs.update({
                "item": item
            })
        response = await callback(*args, **kwargs) if asyncio.iscoroutinefunction(callback) else callback(*args, **kwargs)
        return response

    return wrapper


def item_name_prefix(callback):
    """
    处理 item 里name字段的前缀 prefix
    :param callback:
    :return:
    """

    @wraps(callback)
    async def wrapper(*args, **kwargs):
        if "item" in kwargs and "auth" in kwargs:
            item = kwargs.get("item")
            auth = kwargs.get("auth")
            if hasattr(item, "prefix"):
                item.prefix = item.prefix or auth.prefix
                if hasattr(item, 'name') and item.name:
                    item.name = "%s@%s" % (item.prefix, item.name.replace("%s@" % item.prefix, ""))
                kwargs.update({
                    "item": item
                })
        response = await callback(*args, **kwargs) if asyncio.iscoroutinefunction(callback) else callback(*args, **kwargs)
        return response

    return wrapper


def item_owns(callback):
    """
    处理 item 里owns字段
    :param callback:
    :return:
    """

    @wraps(callback)
    async def wrapper(*args, **kwargs):
        if "item" in kwargs and "auth" in kwargs:
            item = kwargs.get("item")
            auth = kwargs.get("auth")
            item.owns = auth.user.username
            kwargs.update({
                "item": item
            })
        response = await callback(*args, **kwargs) if asyncio.iscoroutinefunction(callback) else callback(*args, **kwargs)
        return response

    return wrapper


class DateEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        else:
            return json.JSONEncoder.default(self, obj)


def log_to_database(session=None, scope=None, methods=None, item=None, auth=None, **kwargs):
    prefix = item.prefix if item and hasattr(item, "prefix") and item.prefix else auth.prefix if auth else None
    if ('post' in methods and scope[:5] == ".post") or 'patch' in methods:
        CrudFunctionLog.init().store(
            session=session,
            item=SchemasStoreUpdateLog(prefix=prefix, scope=scope, methods=",".join(methods), data=json.loads(json.dumps(item.dict(exclude_unset=True), cls=DateEncoder)),
                                       username=auth.user.username),
            close=False)

    if 'delete' in methods:
        CrudFunctionLog.init().store(
            session=session,
            item=SchemasStoreUpdateLog(prefix=prefix, scope=scope, methods=",".join(methods), data=kwargs.get('uuids'), username=auth.user.username),
            close=False)


def route(path: str, module: str, router: APIRouter, methods: Optional[List[str]] = None, **kwargs):
    """
    重写路由修饰器
    :param path:
    :param module:
    :param router:
    :param methods:
    :param kwargs:
    :return:
    """

    if module:
        kwargs.update({"name": "{} {}".format(" ".join(methods), module)})

    def wrap(callback):
        @wraps(callback)
        @router.api_route(path=path.replace("_M_", module), methods=[method.upper() for method in methods], **kwargs)
        async def wrapper(*_args, **_kwargs):
            log_to_database(scope=path.replace("_M_", module), methods=methods, **_kwargs) or None
            return await callback(*_args, **_kwargs) if asyncio.iscoroutinefunction(callback) else callback(*_args, **_kwargs)

        return wrapper

    return wrap



from typing import List

from fastapi import APIRouter, Depends, Security
from sqlalchemy.orm import Session

from yao.db import session as _session
from yao.depends import model_screen_params, model_post_screen_params, auth_user, route, item_owns, item_name_prefix
from yao.schema import ModelScreenParams, Schemas, SchemasError
from yao.function.model import function_queue_name as name
from yao.function.queue.crud import Crud
from yao.function.queue.schema import SchemasPaginateItem, SchemasParams, SchemasResponse, SchemasStoreUpdate
from yao.function.user.schema import SchemasFunctionScopes

from app.advertising.application.crud import Crud as APPLICATIONSCrud

router = APIRouter(tags=[name.replace('.', ' ').title()])

role_scopes = [name, ]


@route('/_M_', module=name, router=router, methods=['get'])
@item_name_prefix
async def post_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_screen_params),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.list" % name])):
    """
    :param session:
    :param params:
    :param auth:
    :return:
    """
    db_model_list = Crud.init().paginate(session=session, where=[('prefix', auth.prefix)], screen_params=params)
    return Schemas(data=SchemasPaginateItem(**db_model_list))


@route('/_M_.post', module=name, router=router, methods=['post'])
@item_name_prefix
async def post_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_post_screen_params),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.list" % name])):
    """
    :param session:
    :param params:
    :param auth:
    :return:
    """
    db_model_list = Crud.init().paginate(session=session, where=[('prefix', auth.prefix)], screen_params=params)
    return Schemas(data=SchemasPaginateItem(**db_model_list))


@route('/_M_.params', module=name, router=router, methods=['get'])
async def params_models(session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.list" % name])):
    """
    :param session:
    :param auth:
    :return:
    """
    data = {
        "applications": APPLICATIONSCrud.init().get(session=session, where=[('prefix', auth.prefix)]),
    }
    return Schemas(data=SchemasParams(**data))


@route('/_M_', module=name, router=router, methods=['post'])
@item_name_prefix
@item_owns
async def store_model(item: SchemasStoreUpdate, session: Session = Depends(_session),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.store" % name])):
    """
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = Crud.init().first(session=session, where=("userName", item.userName))
    if db_model is not None:
        return SchemasError(message="数据已经存在！")
    bool_model = Crud.init().store(session=session, item=item)
    return Schemas(data=SchemasResponse(**bool_model.to_dict()))


@route("/_M_/{uuid}", module=name, router=router, methods=['patch'])
@item_name_prefix
async def update_patch_model(uuid: str, item: SchemasStoreUpdate, session: Session = Depends(_session),
                             auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.update" % name])):
    """
    :param uuid:
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = Crud.init().first(session=session, uuid=uuid)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    if item.queue_status == 0:
        item.start_at = None
        item.stop_at = None
        item.progress = None
        item.progress_text = None
        item.retry = 0
    Crud.init().update(session=session, uuid=uuid, item=item, event=True, exclude_unset=True)
    return Schemas()


@route("/_M_", module=name, router=router, methods=['delete'])
async def delete_models(uuids: List[str], session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.delete" % name])):
    """
    :param uuids:
    :param session:
    :param auth:
    :return:
    """
    bool_model = Crud.init().delete(session=session, uuids=uuids, event=True)
    return Schemas(data=bool_model)



from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel, validator

from yao.schema import Schemas, SchemasPaginate
from yao.function.permission.schema import SchemasFunctionResponse, SchemasFunctionMenuMiniResponse
from yao.function.appointment.schema import SchemasFunctionMiniAppointmentResponse
from yao.function.company.schema import SchemasFunctionResponse as companySchemasFunctionResponse


class SchemasFunctionUserResponse(BaseModel):
    """授权用户 返回"""
    uuid: Optional[str] = None
    prefix: Optional[str] = None
    username: Optional[str] = None
    user_phone: Optional[str] = None
    permissions: Optional[List[SchemasFunctionResponse]] = None
    appointments: Optional[List[SchemasFunctionMiniAppointmentResponse]] = None
    children: Optional[List['SchemasFunctionUserResponse']] = None
    children_ids: Optional[list] = None
    available: Optional[bool] = True
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态

    class Config:
        orm_mode = True


class SchemasFunctionUserMiniResponse(BaseModel):
    """授权用户 返回"""
    uuid: Optional[str] = None
    prefix: Optional[str] = None
    username: Optional[str] = None
    user_phone: Optional[str] = None
    available: Optional[bool] = True

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态

    class Config:
        orm_mode = True


class SchemasPaginateItem(SchemasPaginate):
    items: List[SchemasFunctionUserResponse]


class SchemasFunctionUserStoreUpdate(BaseModel):
    """创建或者更新后台账户信息"""
    prefix: Optional[str] = None
    parent_id: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    user_phone: Optional[str] = None
    available: Optional[bool] = True
    permissions: Optional[List[str]] = None
    appointments: Optional[List[str]] = None
    children_ids: Optional[list] = None
    auth_data: Optional[dict] = None


class SchemasFunctionUserSafeUpdate(BaseModel):
    """创建或者更新后台账户信息"""
    password: Optional[str] = None
    user_phone: Optional[str] = None


class SchemasFunctionUser(BaseModel):
    """解析加密字段"""
    prefix: Optional[str] = None
    sub: Optional[str] = None
    user_id: Optional[int] = 0
    exp: Optional[int] = 0
    scopes: List[str] = []


class _SchemasFunctionUserMiniResponse(SchemasFunctionUserMiniResponse):
    appointments: Optional[list] = None


class SchemasFunctionScopes(BaseModel):
    """验证授权后"""
    prefix: Optional[str] = None
    user: Optional[_SchemasFunctionUserMiniResponse] = None
    scopes: Optional[List[str]] = []
    children_ids: Optional[list] = None


class SchemasLogin(BaseModel):
    """登录返回 token 兼容api"""
    access_token: Optional[str] = None
    token_type: Optional[str] = None


class SchemasLoginResponse(SchemasLogin):
    """登录返回 token"""
    data: Optional[SchemasLogin] = None


class SchemasFunctionUserBriefly(BaseModel):
    """授权用户简要 返回"""
    username: Optional[str] = None
    user_phone: Optional[str] = None
    available: Optional[bool] = True

    class Config:
        orm_mode = True


class SchemasFunctionUserAndScopes(BaseModel):
    """获取授权返回"""
    prefix: Optional[str] = None
    user: Optional[SchemasFunctionUserBriefly] = None
    scopes: Optional[list] = None


class SchemasFunctionUserMeStatusResponse(Schemas):
    """登录授权用户 状态返回"""
    data: SchemasFunctionUserAndScopes


class SchemasParams(BaseModel):
    """参数"""
    appointments: Optional[List[SchemasFunctionMiniAppointmentResponse]] = None
    permissions: Optional[List[SchemasFunctionMenuMiniResponse]] = None
    companies: Optional[List[companySchemasFunctionResponse]] = None


class SchemasBindAuth(BaseModel):
    type: Optional[str] = None
    code: Optional[str] = None
    scene: Optional[str] = None



from typing import List, Optional

from fastapi import APIRouter, Depends, Security, HTTPException, status
from fastapi.responses import HTMLResponse
from fastapi.security import OAuth2PasswordRequestForm
from sqlalchemy.orm import Session

from config import OAUTH_ACCESS_TOKEN_EXPIRE_MINUTES, OAUTH_LOGIN_SCOPES, DEFAULT_FUNCTION_COMPANY, OAUTH_ADMIN_USERS, OAUTH_SECRET_KEY, OAUTH_ALGORITHM, \
    OAUTH_TOKEN_URI, OAUTH_SCOPES_URI, OAUTH_ME_URI
from yao.db import session as _session
from yao.depends import model_screen_params, model_post_screen_params, auth_user
from yao.helpers import token_access_token, token_verify_password
from yao.schema import Schemas, SchemasError, ModelScreenParams
from yao.function.model import function_user_name as name
from yao.function.user.crud import CrudFunctionUser
from yao.function.user.schema import SchemasFunctionScopes, SchemasLoginResponse, SchemasLogin, SchemasFunctionUserMeStatusResponse, SchemasPaginateItem, SchemasParams, \
    SchemasFunctionUserResponse, SchemasFunctionUserStoreUpdate, SchemasFunctionUserSafeUpdate, SchemasBindAuth

router = APIRouter(tags=[name.replace('.', ' ').title()])

user_scopes = [name, ]


def authenticate_user(session: Session, username: str, password: str):
    """
    验证用户信息
    :param session:
    :param username:
    :param password:
    :return:
    """
    user = CrudFunctionUser.init().first(session=session, where=[("username", username), ("available", True)])
    if not user or not token_verify_password(plain_password=password, hashed_password=user.password):
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail="账号或者密码不正确！")
    return user


def token_authenticate_access_token(session, username: str, password: str, scopes: list) -> str:
    """
    认证用户且生成用户token
    :param session:
    :param username:
    :param password:
    :param scopes:
    :return:
    """
    from datetime import timedelta
    user = authenticate_user(session=session, username=username, password=password)
    access_token_expires = timedelta(minutes=int(OAUTH_ACCESS_TOKEN_EXPIRE_MINUTES))
    scopes = scopes + [OAUTH_LOGIN_SCOPES]
    """处理用户拥有的权限"""
    # Todo user.roles user.permission
    if user.username and user.username.split("@")[0] == DEFAULT_FUNCTION_COMPANY.get("prefix_name") and user.username.split("@")[1] in OAUTH_ADMIN_USERS:
        from yao.function.permission.crud import CrudFunctionPermission
        permissions = CrudFunctionPermission.init().get(session=session)
        scopes = scopes + [permission.scope for permission in permissions]
    else:
        # 获取用户权限
        scopes = scopes + [permission.scope for permission in user.permissions]
        # 获取用户角色权限
        scopes = scopes + [scope for appointment in user.appointments for scope in (appointment.scopes.split(' '))]
    """处理用户拥有的权限"""
    return token_access_token(
        data={"sub": user.username, "user_id": user.id, "prefix": user.prefix, "scopes": list(set(scopes))},
        key=OAUTH_SECRET_KEY,
        algorithm=OAUTH_ALGORITHM,
        expires_delta=access_token_expires
    )


@router.post(OAUTH_TOKEN_URI)
async def login_for_access_token(session: Session = Depends(_session), form_data: OAuth2PasswordRequestForm = Depends()):
    """
    获取登录授权:
    - **form_data**: 登录数据
    """
    access_token = token_authenticate_access_token(
        session=session,
        username=form_data.username,
        password=form_data.password,
        scopes=form_data.scopes
    )
    return SchemasLoginResponse(data=SchemasLogin(access_token=access_token, token_type="bearer"), access_token=access_token, token_type="bearer")


@router.get(OAUTH_SCOPES_URI)
async def get_scopes(auth: SchemasFunctionScopes = Security(auth_user)):
    """
    获取登录授权:
    """
    return Schemas(data=auth)


@router.get(OAUTH_ME_URI)
async def get_me(auth: SchemasFunctionScopes = Security(auth_user)):
    """
    获取登录授权的用户信息:
    """
    return Schemas(data=auth.user)


@router.patch(OAUTH_ME_URI)
async def patch_me(item: SchemasFunctionUserSafeUpdate, session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user)):
    """
    更新登录授权用户的信息:
    """
    bool_model = CrudFunctionUser.init(session=session).update(uuid=auth.user.uuid, item=item)
    return Schemas(data=SchemasFunctionUserResponse(**bool_model.to_dict()))


@router.get('/{}'.format(name), name="get {}".format(name))
async def get_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_screen_params),
                     auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.list" % name])):
    """
    获取授权用户列表
    - **:param session**:
    - **:param params**:
    - **:param auth**:
    - **:return**:
    """
    if auth.user.username and auth.user.username.split("@")[0] == DEFAULT_FUNCTION_COMPANY.get("prefix_name") and auth.user.username.split("@")[1] in OAUTH_ADMIN_USERS:
        db_model_list = CrudFunctionUser.init().paginate(session=session, where=[("parent_id", None) if len(params.where) == 0 else None], screen_params=params)
    else:
        db_model_list = CrudFunctionUser.init().paginate(session=session, where=[('prefix', auth.prefix), ("parent_id", None) if len(params.where) == 0 else None],
                                                         screen_params=params)
    return Schemas(data=SchemasPaginateItem(**db_model_list))


@router.post('/{}.post'.format(name), name="post {}".format(name))
async def post_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_post_screen_params),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.list" % name])):
    """
    获取授权用户列表
    - **:param session**:
    - **:param params**:
    - **:param auth**:
    - **:return**:
    """
    db_model_list = CrudFunctionUser.init().paginate(session=session, screen_params=params)
    return Schemas(data=SchemasPaginateItem(**db_model_list))


@router.get('/{}.params'.format(name), name="get {}".format(name))
async def params_models(session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.list" % name])):
    """
    :param session:
    :param auth:
    :return:
    """
    from yao.function.permission.crud import CrudFunctionPermission
    from yao.function.appointment.crud import CrudFunctionAppointment
    if auth.user.username and auth.user.username.split("@")[0] == DEFAULT_FUNCTION_COMPANY.get("prefix_name") and auth.user.username.split("@")[1] in OAUTH_ADMIN_USERS:
        from yao.function.company.crud import CrudFunctionCompany
        data = {
            "appointments": CrudFunctionAppointment.init().get(session=session),
            "permissions": CrudFunctionPermission.init().get_tree(session=session, json=True),
            "companies": CrudFunctionCompany.init().get(session=session)
        }
    else:
        has_appointment_uuids = [appointment.uuid for appointment in auth.user.appointments]
        data = {
            "appointments": CrudFunctionAppointment.init().get(session=session, where=[("__or", [("prefix", auth.prefix), ("uuid", "in", has_appointment_uuids)])]),
            "permissions": CrudFunctionPermission.init().get_tree(session=session, where=[("scope", "in", auth.scopes)], json=True)
        }
    return Schemas(data=SchemasParams(**data))


@router.get('/{}/{{uuid}}'.format(name), name="get {}".format(name))
async def get_model(uuid: str, session: Session = Depends(_session),
                    auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.get" % name])):
    """
    :param uuid:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionUser.init(session=session).first(uuid=uuid)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    return Schemas(data=SchemasFunctionUserResponse(**db_model.to_dict()))


@router.post('/{}'.format(name), name="get {}".format(name))
async def store_model(item: SchemasFunctionUserStoreUpdate, session: Session = Depends(_session),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.store" % name])):
    """
    :param item:
    :param session
    :param auth:
    :return:
    """
    item.prefix = item.prefix or auth.prefix
    item.username = "%s@%s" % (item.prefix, item.username.replace("%s@" % item.prefix, ""))
    db_model = CrudFunctionUser.init().first(session=session, where=("username", item.username))
    if db_model is not None:
        return SchemasError(message="数据已经存在！")
    bool_model = CrudFunctionUser.init().store(session=session, item=item)
    return Schemas(data=SchemasFunctionUserResponse(**bool_model.to_dict()))


@router.put("/{}/{{uuid}}".format(name), name="update {}".format(name))
async def update_put_model(uuid: str, item: SchemasFunctionUserStoreUpdate, session: Session = Depends(_session),
                           auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.update" % name])):
    """
    :param uuid:
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionUser.init(session=session).first(uuid=uuid)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    item.prefix = item.prefix or auth.prefix
    item.username = "%s@%s" % (auth.prefix, item.username)
    CrudFunctionUser.init(session=session).update(uuid=uuid, item=item, exclude_unset=False)
    return Schemas()


@router.patch("/{}/{{uuid}}".format(name), name="update {}".format(name))
async def update_patch_model(uuid: str, item: SchemasFunctionUserStoreUpdate, session: Session = Depends(_session),
                             auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.update" % name])):
    """
    :param uuid:
    :param item:
    :param session:
    :param auth:
    :return:
    """
    item.prefix = item.prefix or auth.prefix
    item.username = "%s@%s" % (item.prefix, item.username.replace("%s@" % item.prefix, ""))
    db_model = CrudFunctionUser.init().first(session=session, uuid=uuid)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    CrudFunctionUser.init().update(session=session, uuid=uuid, item=item, exclude_unset=True, event=True, close=True)
    return Schemas()


@router.delete("/{}/{{uuid}}".format(name), name="delete {}".format(name))
async def delete_model(uuid: str, session: Session = Depends(_session),
                       auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.delete" % name])):
    """
    :param uuid:
    :param session:
    :param auth:
    :return:
    """
    bool_model = CrudFunctionUser.init(session=session).delete(session=session, uuid=uuid)
    return Schemas(data=bool_model)


@router.delete("/{}".format(name), name="deletes {}".format(name))
async def delete_models(pks: List[int], session: Session = Depends(_session),
                        auth: SchemasFunctionScopes = Security(auth_user, scopes=user_scopes + ["%s.delete" % name])):
    """
    :param pks:
    :param session:
    :param auth:
    :return:
    """
    bool_model = CrudFunctionUser.init(session=session).delete(pks=pks)
    return Schemas(data=bool_model)


@router.post('/{}.bind.auth'.format(name), name="post {}".format(name))
async def bind_auth(item: SchemasBindAuth, session: Session = Depends(_session)):
    """
    :param item:
    :param session
    :return:
    """
    try:
        from config import PROGRAMAPPID, PROGRAMAPPSECRET
    except:
        return SchemasError(data="配置出错！")
    from yao.wxamp.base import AMP
    try:
        amp = AMP(PROGRAMAPPID, PROGRAMAPPSECRET)
        res = amp.code2session(item.code)
        openid = res.get("openid", None)
        if item.scene and openid:
            db_model = CrudFunctionUser.init().first(session=session, uuid=item.scene)
            auth_data = db_model.auth_data if db_model and type(db_model.auth_data) is dict else {}
            auth_data.update({
                item.type: {"openid": openid}
            })
            _item = SchemasFunctionUserStoreUpdate(auth_data=auth_data)
            CrudFunctionUser.init().update(session=session, uuid=item.scene, item=_item, exclude_unset=True, event=True, close=True)
            return Schemas(data="已经成功获取到授权，可以关闭当前窗口！")
    except:
        return SchemasError(data="授权出错！")


@router.get('/auth/mp/{user_uuid}', name="get {}".format(name))
async def mp_auth(user_uuid: str, session: Session = Depends(_session)):
    """
    :param user_uuid
    :param session
    :return:
    """
    from config import HOME_URL
    try:
        from config import MPAPPID, MPAPPSECRET
    except:
        return SchemasError(data="配置出错！")
    from yao.wxmp.base import MP
    try:
        mp = MP(MPAPPID, MPAPPSECRET)
        redirect_url = "%sapi/auth/mp_callback" % HOME_URL
        return mp.get_or_to_auth_url(redirect_uri=redirect_url, state=user_uuid, fastapi_return=True)
    except:
        return SchemasError(data="授权出错！")


@router.get('/auth/mp_callback', name="get {}".format(name))
async def callback_auth(code, state: str, session: Session = Depends(_session)):
    """
    :param code
    :param state user_uuid
    :param session
    :return:
    """
    try:
        from config import MPAPPID, MPAPPSECRET
    except:
        return SchemasError(data="配置出错！")
    from yao.wxmp.base import MP
    try:
        mp = MP(MPAPPID, MPAPPSECRET)
        res = mp.code_to_openid(code=code)
        openid = res.get("openid", None)

        if state and openid:
            db_model = CrudFunctionUser.init().first(session=session, uuid=state)
            auth_data = db_model.auth_data if db_model and type(db_model.auth_data) is dict else {}
            auth_data.update({
                "mp": {"openid": openid}
            })
            _item = SchemasFunctionUserStoreUpdate(auth_data=auth_data)
            CrudFunctionUser.init().update(session=session, uuid=state, item=_item, exclude_unset=True, close=True)
            content = """
                    <html>
                    <head>
                        <meta charset="UTF-8">
                        <meta name="viewport" content="width=device-width, user-scalable=no, initial-scale=1.0, maximum-scale=1.0, minimum-scale=1.0">
                        <meta http-equiv="X-UA-Compatible" content="ie=edge">
                        <title>授权成功!</title>
                    </head>
                    <body>
                    <h1 style="text-align: center;margin:100px auto;width:100%;color:green">已经成功获取到授权，可以关闭当前窗口！</h1>
                    </body>
                    </html>
                    """
            return HTMLResponse(content=content, status_code=200)
    except:
        pass
    content = """
            <html>
            <head>
                <meta charset="UTF-8">
                <meta name="viewport" content="width=device-width, user-scalable=no, initial-scale=1.0, maximum-scale=1.0, minimum-scale=1.0">
                <meta http-equiv="X-UA-Compatible" content="ie=edge">
                <title>授权出错!</title>
            </head>
            <body>
            <h1 style="text-align: center;margin:100px auto;width:100%;color:red">授权出错！</h1>
            </body>
            </html>
            """
    return HTMLResponse(content=content, status_code=200)



from datetime import datetime
from typing import Optional, List, Union

from pydantic import BaseModel, validator

from yao.schema import SchemasPaginate


class SchemasFunctionResponse(BaseModel):
    """模型 返回"""
    uuid: Optional[str] = None
    scope: Optional[str] = None  # scope
    methods: Optional[str] = None  # methods
    data: Optional[Union[dict, list]] = None  # Data
    username: Optional[str] = None  # 权限

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    @validator('username')
    def p_name(cls, username: str):
        return username.split("@", 1)[1]

    class Config:
        orm_mode = True


class SchemasFunctionStoreUpdate(BaseModel):
    """模型 提交"""
    prefix: Optional[str] = None
    uuid: Optional[str] = None
    scope: Optional[str] = None  # scope
    methods: Optional[str] = None  # methods
    data: Optional[Union[dict, list]] = None  # Data
    username: Optional[str] = None  # 权限

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态


class SchemasFunctionPaginateItem(SchemasPaginate):
    items: List[SchemasFunctionResponse]


class SchemasFunctionParams(BaseModel):
    pass



from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel

from yao.schema import SchemasPaginate, SchemaPrefixNames
from app.stats.schema import SchemasPrefixOwns
from app.advertising.application.schema import SchemasResponse as SchemasApplicationResponse


class SchemasResponse(SchemaPrefixNames):
    """模型 返回"""
    uuid: Optional[str] = None
    username: Optional[str] = None  # 用户
    priority: Optional[int] = None  # 优先级
    scope: Optional[str] = None  # scope
    data: Optional[dict] = None  # Data
    key: Optional[str] = None  # 去重Key
    start_at: Optional[datetime] = None  # 开始时间
    stop_at: Optional[datetime] = None  # 结束时间
    progress: Optional[str] = None  # 进度条
    progress_text: Optional[str] = None  # 进度条
    retry: Optional[int] = None  # 重试次数
    queue_status: Optional[int] = None  # 状态

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True


class SchemasStoreUpdate(SchemasPrefixOwns):
    """模型 提交"""
    username: Optional[str] = None  # 用户
    priority: Optional[int] = None  # 优先级
    scope: Optional[str] = None  # scope
    data: Optional[dict] = None  # Data
    key: Optional[str] = None  # 去重Key
    start_at: Optional[datetime] = None  # 开始时间
    stop_at: Optional[datetime] = None  # 结束时间
    progress: Optional[str] = None  # 进度条
    progress_text: Optional[str] = None  # 进度条
    retry: Optional[int] = None  # 重试次数
    queue_status: Optional[int] = None  # 状态 0未运行 1成功 2失败

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态


class SchemasPaginateItem(SchemasPaginate):
    items: List[SchemasResponse]


class SchemasParams(BaseModel):
    applications: Optional[List[SchemasApplicationResponse]]


class SchemasQueueAuth(BaseModel):
    prefix: Optional[str] = None
    username: Optional[str] = None



import json
import os
import requests
import filetype

from fastapi import APIRouter, Depends, Security, UploadFile, File, Form, Request
from sqlalchemy.orm import Session
from typing import Optional

from yao.db import session as _session
from yao.depends import auth_user, route, item_prefix
from yao.schema import Schemas, SchemasError
from yao.function.model import function_annex_name as name
from yao.function.annex.crud import CrudFunctionAnnexe
from yao.function.annex.schema import SchemasFunctionAnnexeResponse, SchemasFunctionAnnexeStoreUpdate, SchemasUpLoadFileResponse, SchemasUpLoadContentFile
from yao.function.user.schema import SchemasFunctionScopes
from config import UPLOAD_DIR
from yao.method import md5_bytes, write_file

router = APIRouter(tags=[name.replace('.', ' ').title()])

role_scopes = [name, ]


@route('/_M_', module=name, router=router, methods=['post'])
async def store_model(file: UploadFile = File(...), session: Session = Depends(_session),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.store" % name])):
    """
    :param file:
    :param session:
    :param auth:
    :return:
    """
    content = await file.read()
    md5 = md5_bytes(content)
    size = len(content)

    db_model = CrudFunctionAnnexe.init().first(session=session, where=[("md5", md5), ("prefix", auth.prefix)])
    if db_model is None:
        path = os.path.join(UPLOAD_DIR, auth.prefix, os.path.split(file.content_type)[0], "{}{}".format(md5, os.path.splitext(file.filename)[-1]))
        bool_res = write_file(path, content)
        if bool_res:
            try:
                from PIL import Image
                img = Image.open(path)
                width, height = img.size
            except:
                width, height = (0, 0)
            CrudFunctionAnnexe.init().store(session=session,
                                            item=SchemasFunctionAnnexeStoreUpdate(
                                                prefix=auth.prefix,
                                                filename=file.filename,
                                                content_type=file.content_type,
                                                md5=md5, path=path, size=size, width=width, height=height))
            db_model = CrudFunctionAnnexe.init().first(session=session, where=("md5", md5))
    return Schemas(data=SchemasFunctionAnnexeResponse(**db_model.to_dict()))


@route('/_M_.ueditor', module=name, router=router, methods=['get'])
@item_prefix
async def get_ueditor_models(action: str, session: Session = Depends(_session)):
    """
    :param action:
    :param session:
    :return:
    """
    # 前后端通信相关的配置,注释只允许使用多行方式 
    defaultConfig = {
        # 上传图片配置项 
        "imageActionName": "uploadimage",  # 执行上传图片的action名称 
        "imageFieldName": "file",  # 提交的图片表单名称
        "imageMaxSize": 2048000,  # 上传大小限制，单位B 
        "imageAllowFiles": [".png", ".jpg", ".jpeg", ".gif", ".bmp"],  # 上传图片格式显示 
        "imageCompressEnable": True,  # 是否压缩图片,默认是True 
        "imageCompressBorder": 1600,  # 图片压缩最长边限制 
        "imageInsertAlign": "none",  # 插入的图片浮动方式 
        "imageUrlPrefix": "",  # 图片访问路径前缀

        # 涂鸦图片上传配置项 
        "scrawlActionName": "uploadscrawl",  # 执行上传涂鸦的action名称 
        "scrawlFieldName": "file",  # 提交的图片表单名称
        "scrawlPathFormat": "/ueditor/php/upload/image/{yyyy}{mm}{dd}/{time}{rand:6}",  # 上传保存路径,可以自定义保存路径和文件名格式 
        "scrawlMaxSize": 2048000,  # 上传大小限制，单位B 
        "scrawlUrlPrefix": "",  # 图片访问路径前缀 
        "scrawlInsertAlign": "none",

        # 截图工具上传 
        "snapscreenActionName": "uploadimage",  # 执行上传截图的action名称 
        "snapscreenPathFormat": "/ueditor/php/upload/image/{yyyy}{mm}{dd}/{time}{rand:6}",  # 上传保存路径,可以自定义保存路径和文件名格式 
        "snapscreenUrlPrefix": "",  # 图片访问路径前缀 
        "snapscreenInsertAlign": "none",  # 插入的图片浮动方式 

        # 抓取远程图片配置 
        "catcherLocalDomain": ["127.0.0.1", "localhost", "img.baidu.com"],
        "catcherActionName": "catchimage",  # 执行抓取远程图片的action名称 
        "catcherFieldName": "source",  # 提交的图片列表表单名称 
        "catcherPathFormat": "/ueditor/php/upload/image/{yyyy}{mm}{dd}/{time}{rand:6}",  # 上传保存路径,可以自定义保存路径和文件名格式 
        "catcherUrlPrefix": "",  # 图片访问路径前缀 
        "catcherMaxSize": 2048000,  # 上传大小限制，单位B 
        "catcherAllowFiles": [".png", ".jpg", ".jpeg", ".gif", ".bmp"],  # 抓取图片格式显示 

        # 上传视频配置 
        "videoActionName": "uploadvideo",  # 执行上传视频的action名称 
        "videoFieldName": "file",  # 提交的视频表单名称
        "videoPathFormat": "/ueditor/php/upload/video/{yyyy}{mm}{dd}/{time}{rand:6}",  # 上传保存路径,可以自定义保存路径和文件名格式 
        "videoUrlPrefix": "",  # 视频访问路径前缀 
        "videoMaxSize": 102400000,  # 上传大小限制，单位B，默认100MB 
        "videoAllowFiles": [
            ".flv", ".swf", ".mkv", ".avi", ".rm", ".rmvb", ".mpeg", ".mpg",
            ".ogg", ".ogv", ".mov", ".wmv", ".mp4", ".webm", ".mp3", ".wav", ".mid"],  # 上传视频格式显示 

        # 上传文件配置 
        "fileActionName": "uploadfile",  # controller里,执行上传视频的action名称 
        "fileFieldName": "file",  # 提交的文件表单名称
        "filePathFormat": "/ueditor/php/upload/file/{yyyy}{mm}{dd}/{time}{rand:6}",  # 上传保存路径,可以自定义保存路径和文件名格式 
        "fileUrlPrefix": "",  # 文件访问路径前缀 
        "fileMaxSize": 51200000,  # 上传大小限制，单位B，默认50MB 
        "fileAllowFiles": [
            ".png", ".jpg", ".jpeg", ".gif", ".bmp",
            ".flv", ".swf", ".mkv", ".avi", ".rm", ".rmvb", ".mpeg", ".mpg",
            ".ogg", ".ogv", ".mov", ".wmv", ".mp4", ".webm", ".mp3", ".wav", ".mid",
            ".rar", ".zip", ".tar", ".gz", ".7z", ".bz2", ".cab", ".iso",
            ".doc", ".docx", ".xls", ".xlsx", ".ppt", ".pptx", ".pdf", ".txt", ".md", ".xml"
        ],  # 上传文件格式显示 

        # 列出指定目录下的图片 
        "imageManagerActionName": "listimage",  # 执行图片管理的action名称 
        "imageManagerListPath": "/ueditor/php/upload/image/",  # 指定要列出图片的目录 
        "imageManagerListSize": 20,  # 每次列出文件数量 
        "imageManagerUrlPrefix": "",  # 图片访问路径前缀 
        "imageManagerInsertAlign": "none",  # 插入的图片浮动方式 
        "imageManagerAllowFiles": [".png", ".jpg", ".jpeg", ".gif", ".bmp"],  # 列出的文件类型 

        # 列出指定目录下的文件 
        "fileManagerActionName": "listfile",  # 执行文件管理的action名称 
        "fileManagerListPath": "/ueditor/php/upload/file/",  # 指定要列出文件的目录 
        "fileManagerUrlPrefix": "",  # 文件访问路径前缀 
        "fileManagerListSize": 20,  # 每次列出文件数量 
        "fileManagerAllowFiles": [
            ".png", ".jpg", ".jpeg", ".gif", ".bmp",
            ".flv", ".swf", ".mkv", ".avi", ".rm", ".rmvb", ".mpeg", ".mpg",
            ".ogg", ".ogv", ".mov", ".wmv", ".mp4", ".webm", ".mp3", ".wav", ".mid",
            ".rar", ".zip", ".tar", ".gz", ".7z", ".bz2", ".cab", ".iso",
            ".doc", ".docx", ".xls", ".xlsx", ".ppt", ".pptx", ".pdf", ".txt", ".md", ".xml"
        ]  # 列出的文件类型 

    }
    if action == "config":
        return defaultConfig


def find_or_create_file(session, content, prefix, filename=None):
    md5 = md5_bytes(content)
    size = len(content)
    file = filetype.guess(content)
    db_model = CrudFunctionAnnexe.init().first(session=session, where=[("md5", md5), ("prefix", prefix)])
    if db_model is None:
        path = os.path.join(UPLOAD_DIR, prefix, os.path.split(file.mime)[0], "{}.{}".format(md5, file.extension))
        bool_res = write_file(path, content)
        if bool_res:
            try:
                from PIL import Image
                img = Image.open(path)
                width, height = img.size
            except:
                width, height = (0, 0)
            filename = filename or "{}.{}".format(md5, file.extension)
            CrudFunctionAnnexe.init().store(
                session=session, item=SchemasFunctionAnnexeStoreUpdate(
                    prefix=prefix, filename=filename, content_type=file.mime,
                    md5=md5, path=path, size=size, width=width, height=height
                )
            )
            db_model = CrudFunctionAnnexe.init().first(session=session, where=("md5", md5))
    return db_model


@route('/_M_.ueditor', module=name, router=router, methods=['post'])
async def store_ueditor_model(action: str, request: Request, file: Optional[UploadFile] = File(default=None), source: list = Form(), session: Session = Depends(_session),
                              auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.store" % name])):
    if action == "uploadimage":
        content = await file.read()
        db_model = find_or_create_file(session=session, content=content, prefix=auth.prefix, filename=file.filename)
        return SchemasUpLoadFileResponse(url=db_model.preview_path, type=db_model.content_type, size=db_model.size, original=db_model.filename, title=db_model.filename)
    elif action == "catchimage":
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.105 Safari/537.36'
        }
        catcher = []
        for _source in source:
            r = requests.get(_source, headers=headers, stream=True)
            if r.status_code == 200:
                content = r.content
                db_model = find_or_create_file(session=session, content=content, prefix=auth.prefix)
                state = "SUCCESS"
                url = db_model.preview_path
            else:
                state = "ERROR_UNKNOWN"
                url = ""
            catcher.append({
                "state": state,
                "url": url,
                "source": _source
            })
        state = "SUCCESS" if bool(len(catcher)) else "ERROR"
        return SchemasUpLoadFileResponse(state=state, list=catcher)


@route('/_M_.content.file', module=name, router=router, methods=['post'])
async def content_file_model(path: str = Form(), file: UploadFile = File(...)):
    """
    留给模板用
    :param path:
    :param file:
    :return:
    """
    content = await file.read()
    bool_res = write_file(path=os.path.join(UPLOAD_DIR, "template", path), content=content)
    return Schemas() if bool_res else SchemasError()



import base64
import hashlib
import json
import os
import random
import re
import socket
import uuid
import zipfile
import datetime


def hashids_encode(ids: str, **kwargs):
    """
    hashids 加密
    :param ids:
    :param kwargs:
    :return:
    """
    from hashids import Hashids
    hashids = Hashids(**kwargs)
    return hashids.encode(ids)


def hashids_decode(string: str, **kwargs):
    """
    hashids 解密
    :param string:
    :param kwargs:
    :return:
    """
    from hashids import Hashids
    hashids = Hashids(**kwargs)
    return hashids.decode(string)


def get_mac_address():
    """
    获取 本机 mac 地址
    :return:
    """
    mac = uuid.UUID(int=uuid.getnode()).hex[-12:]
    return ":".join([mac[e:e + 2] for e in range(0, 11, 2)])


def get_host_name():
    """
    获取主机名
    :return:
    """
    return socket.getfqdn(socket.gethostname())


def get_host_ip():
    """
    获取IP地址
    :return:
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    finally:
        s.close()
    return ip


def get_client_ip(request):
    """
    获取请求的ip
    :param request:
    :return:
    """
    headers = request.headers
    addr = headers.get('http_x_forwarded_for', None)
    if not addr:
        addr = headers.get('http_client_ip', None)
    if not addr:
        addr = headers.get('remote_addr', None)
    if not addr:
        addr = headers.get('remote-host', None)
    if not addr:
        addr = headers.get('x-real-ip', None)
    if not addr:
        addr = headers.get('x-forwarded-for', None)
    if not addr:
        addr = headers.get('x-forwarded-for', None)
    return addr or None


def md5_bytes(string: bytes):
    """
    md5 字符串
    :param string:
    :return:
    """
    md5hash = hashlib.md5(string)
    return md5hash.hexdigest()


def md5_string(string: str, encoding="utf8"):
    """
    md5 字符串
    :param string:
    :param encoding:
    :return:
    """
    return md5_bytes(string.encode(encoding))


def md5_file(file):
    """
    md5 文件
    :param file:
    :return:
    """
    m = hashlib.md5()
    with open(file, 'rb') as obj:
        while True:
            data = obj.read(4096)
            if not data:
                break
            m.update(data)

    return m.hexdigest()


def str_bytes(string: str, **kwargs):
    """
    字符串转字节
    :param string:
    :param kwargs:
    :return:
    """
    return bytes(string, **kwargs)


def bytes_str(b: bytes, **kwargs):
    """
    字节转字符串
    :param b:
    :param kwargs:
    :return:
    """
    return str(b, **kwargs)


def plural(word: str):
    """
    字符串转复制
    :param word:
    :return:
    """
    if word.endswith('y'):
        return word[:-1] + "ies"
    elif word[-1] in 'sx' or word[-2:] in ["sh", "ch"]:
        return word + "es"
    elif word.endswith('an'):
        return word[:-2] + "en"
    else:
        return word + "s"


def write_file(path, content, mode="wb", **kwargs):
    """
    写入文件
    :param path:
    :param content:
    :param mode:
    :param kwargs:
    :return:
    """
    dir = os.path.dirname(path)
    if not os.path.exists(dir):
        os.makedirs(dir)
    if not os.path.isdir(path):
        with open(path, mode, **kwargs) as f:
            f.write(content)
        return True
    else:
        return False


def zip_files(files, zip_name):
    """
    压缩文件
    """
    if not os.path.exists(os.path.dirname(zip_name)):  # os模块判断并创建
        os.makedirs(os.path.dirname(zip_name))
    zip = zipfile.ZipFile(zip_name, 'w', zipfile.ZIP_DEFLATED)
    for file in files:
        zip.write(file)
    zip.close()
    return True


def count(arr: list):
    """计算数据元素总和"""
    total = 0
    for num in arr:
        total += num
    return total


def probability_extract(collect: list, key: str = "probability", type=None):
    """
    概率抽取
    :param collect: 抽取list
    :param key: 概率 键
    :param type: 概率 键
    :return:
    """

    def get(ob, key, default):
        if type == "dict":
            return ob.get(key, default)
        else:
            return getattr(ob, key, default)

    res = None
    sum = count([get(ob, key, 0) for ob in collect])
    if sum > 0:
        randint = random.randint(1, sum)
        for ob in collect:
            probability = get(ob, key, 0)
            if randint > probability:
                randint -= probability
            else:
                res = ob
                break
        return res
    return None


def get_file_content(path: str, mode="rb", **kwargs):
    """获取文件内容"""
    if os.path.isfile(path):
        with open(path, mode, **kwargs) as f:
            return f.read()
    return bytes()


def file_to_base64(path: str):
    """文件转base64"""
    import filetype
    content = get_file_content(path)
    base = base64.b64encode(content).decode()
    type = filetype.guess(path)
    return "data:%s;base64,%s" % (type.mime, base)


def filter_content(content: str):
    """过滤内容"""
    return re.sub(r'\s+', ' ', content)  # 去掉换行 和 多余空格


def convert(one_string, space_character="_"):
    """
    字符串转成驼峰
    :param one_string: 输入的字符串
    :param space_character: 字符串的间隔符，以其做为分隔标志
    :return:
    """
    string_list = str(one_string).split(space_character)  # 将字符串转化为list
    first = string_list[0].lower()
    others = string_list[1:]
    others_capital = [word.capitalize() for word in others]  # str.capitalize():将字符串的首字母转化为大写
    others_capital[0:0] = [first]
    hump_string = ''.join(others_capital)  # 将list组合成为字符串，中间无连接符。
    return hump_string


def convert_dict_key(data: dict, space_character="_"):
    """
    字典键转驼峰
    :param data:
    :param space_character:
    :return:
    """
    return {convert(key, space_character=space_character): value for key, value in data.items()}


def convert_key(data, space_character="_"):
    """
    字典键转驼峰
    :param data:
    :param space_character:
    :return:
    """
    if type(data) is dict:
        return {convert(key, space_character=space_character): convert_key(value) if type(value) is list or type(value) is dict else value for key, value in data.items()}
    elif type(data) is list:
        return [convert_key(datum) if type(datum) is dict else datum for datum in data]
    return data


def aes_encrypt(data, key="0000000000000000", iv="0000000000000000"):
    """
    AES的ECB模式加密方法
    :param data:被加密字符串（明文）
    :param key: 密钥
    :param iv: 密斯偏移量
    :return:密文
    """
    from Crypto.Cipher import AES
    BLOCK_SIZE = 16  # Bytes
    pad = lambda s: s + (BLOCK_SIZE - len(s) % BLOCK_SIZE) * chr(BLOCK_SIZE - len(s) % BLOCK_SIZE)
    key = key.encode('utf8')
    data = pad(data)
    cipher = AES.new(key, AES.MODE_EAX, iv.encode('utf8'))
    # 加密后得到的是bytes类型的数据，使用Base64进行编码,返回byte字符串
    result = cipher.encrypt(data.encode('utf8'))
    return base64.b64encode(result).decode('utf8')


def aes_decrypt(data, key="0000000000000000", iv="0000000000000000"):
    """
    :param data: 加密后的数据（密文）
    :param key: 密钥
    :param iv: 密斯偏移量
    :return:明文
    """
    from Crypto.Cipher import AES
    unpad = lambda s: s[:-ord(s[len(s) - 1:])]
    key = key.encode('utf8')
    data = base64.b64decode(data)
    cipher = AES.new(key, AES.MODE_EAX, iv.encode('utf8'))
    return unpad(cipher.decrypt(data)).decode('utf8')


def filter_None_key(data):
    """
    过滤值为None的键
    :param data:
    :return:
    """
    if type(data) is dict:
        return {key: value for key, value in data.items() if value is not None}
    elif type(data) is list:
        return [filter_None_key(datum) for datum in data]
    return data


def get_attr(object, name: str, default=None):
    """
    获取对象属性 多层 以.分割
    """
    _name = name.split('.')
    for name in _name:
        if type(object) is dict:
            object = object.get(name, default)
        else:
            object = getattr(object, name, default)
    if object is None:
        object = ""
    if object is True:
        object = 1
    if object is False:
        object = 0
    if type(object) is dict:
        object = json.dumps(object)
    return object


def export_file(sheet_name: str, export_name: str, col_items: dict = {}, db_list: list = [], is_header=True):
    """
    导出文件
    """
    from openpyxl import Workbook
    wb = Workbook()
    ws = wb.active
    ws.title = sheet_name
    is_header and ws.append([datum for datum in col_items.values()])
    # [ws.append([str(get_attr(db_obj, key, "")) for key in col_items]) for db_obj in db_list]
    for db_obj in db_list:
        if type(db_obj) is list:
            ws.append(db_obj)
        else:
            _list = []
            for key in col_items:
                _list.append(get_attr(db_obj, key, ""))
            ws.append(_list)

    if not os.path.exists(os.path.dirname(export_name)):
        os.makedirs(os.path.dirname(export_name))
    return wb.save(export_name)


def export_files(export_name: str, sheet_data: list):
    """
    导出文件
    """
    from openpyxl import Workbook
    wb = Workbook()
    for key, sheet in enumerate(sheet_data):
        ws = wb.create_sheet(sheet.get("sheet_name"), key)
        header_label = sheet.get("col_items", {})
        if sheet.get("is_label", True):
            ws.append([datum for datum in header_label.values()])
        for db_obj in sheet.get("db_list", []):
            if type(db_obj) is list:
                ws.append(db_obj)
            else:
                _list = []
                for key in header_label.keys():
                    _list.append(get_attr(db_obj, key, ""))
                ws.append(_list)

        if not os.path.exists(os.path.dirname(export_name)):
            os.makedirs(os.path.dirname(export_name))
    return wb.save(export_name)


def export_file_to_dict(sheet_name: str, export_name: str, col_items: dict, db_list: list):
    """
    导出文件
    """
    from openpyxl import Workbook
    wb = Workbook()
    ws = wb.active
    ws.title = sheet_name
    ws.append([datum for datum in col_items.values()])
    [ws.append([str(db_obj[key] if key in db_obj else "") for key in col_items]) for db_obj in db_list]
    if not os.path.exists(os.path.dirname(export_name)):
        os.makedirs(os.path.dirname(export_name))
    return wb.save(export_name)


def import_file(file: str, sheet: int = 0) -> list:
    """读取excel文件内容"""
    from openpyxl import load_workbook
    wb = load_workbook(file)
    # 获取所有表格(worksheet)的名字
    sheets = wb.sheetnames
    # 第一个表格的名称
    sheet_first = sheets[sheet]
    # 获取特定的worksheet
    ws = wb[sheet_first]
    # 获取表格所有行和列，两者都是可迭代的
    rows = ws.rows
    content = []
    # 迭代所有的行
    for row in rows:
        # line = []
        # for col in row:
        #     line.append((type(col.value) ===str ? col.value.strip(): col.value ))
        line = [(col.value.strip() if type(col.value) == str else col.value) for col in row]
        content.append(line)
    return content


def export_demo(sheet_name: str, export_name: str, col_items: dict, db_list: list):
    """https://www.cnblogs.com/MDD-Blog/p/14187228.html"""
    from openpyxl import Workbook
    wb = Workbook()
    # ws = wb.active
    # ws.title = sheet_name
    ws = wb.create_sheet(sheet_name, 0)
    ws.append([datum for datum in col_items.values()])
    [ws.append([str(get_attr(db_obj, key, "")) for key in col_items]) for db_obj in db_list]
    if not os.path.exists(os.path.dirname(export_name)):
        os.makedirs(os.path.dirname(export_name))
    return wb.save(export_name)


def string_replace(content: str, str_dict: dict) -> str:
    """
    字典替换
    :param content:
    :param str_dict:
    :return:
    """
    for item, value in str_dict.items():
        content = content.replace(item, value)
    return content


def string_re_replace(content: str, str_dict: dict) -> str:
    """
    字典替换
    :param content:
    :param str_dict:
    :return:
    """
    for item, value in str_dict.items():
        content = re.sub(item, value, content)
    return content


def get_date_of_last_month(form="%Y-%m-%d"):
    """
    获取上月开始结束日期
    :param form 返回值显示格式
    :return: str，date tuple
    """
    today = datetime.date.today()
    year = today.year
    month = today.month
    if month == 1:
        begin_of_last_month = datetime.date(year - 1, 12, 1).strftime(form)
    else:
        begin_of_last_month = datetime.date(year, month - 1, 1).strftime(form)
    end_of_last_month = (datetime.date(year, month, 1) + datetime.timedelta(-1)).strftime(form)
    return begin_of_last_month, end_of_last_month


def get_qrcode(data, path, version=1, box_size=10, border=5, img_fill_color="black", img_back_color="white"):
    """
    生成二维码
    Args:
        data:
        path:
        version:
        box_size:
        border:
        img_fill_color:
        img_back_color:

    Returns:

    """
    from qrcode import QRCode
    # 创建QRCode对象
    qr = QRCode(version=version, box_size=box_size, border=border)
    # 设置二维码数据
    qr.add_data(data)
    # 填充数据并生成二维码
    qr.make(fit=True)
    img = qr.make_image(fill_color=img_fill_color, back_color=img_back_color)
    # 保存二维码图片
    img.save(path)
    return True



from typing import List

from fastapi import APIRouter, Depends, Security
from sqlalchemy.orm import Session

from yao.db import session as _session
from yao.depends import model_screen_params, model_post_screen_params, auth_user
from yao.schema import ModelScreenParams, Schemas, SchemasError
from yao.function.model import function_permission_name as name
from yao.function.permission.crud import CrudFunctionPermission
from yao.function.permission.schema import SchemasFunctionPaginateItem, SchemasFunctionResponse, SchemasFunctionStoreUpdate, \
    SchemasFunctionMenuStatusResponse
from yao.function.user.schema import SchemasFunctionScopes

router = APIRouter(tags=[name.replace('.', ' ').title()])

permission_scopes = [name, ]


@router.get('/{}'.format(name), name="get {}".format(name))
async def get_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_screen_params),
                     auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.list" % name])):
    """
    :param session:
    :param params:
    :param auth:
    :return:
    """
    db_model_list = CrudFunctionPermission.init().paginate(session=session, tree=True)
    return Schemas(data=SchemasFunctionPaginateItem(**db_model_list))


@router.post('/{}.post'.format(name), name="post {}".format(name))
async def post_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_post_screen_params),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.list" % name])):
    """
    :param session:
    :param params:
    :param auth:
    :return:
    """
    db_model_list = CrudFunctionPermission.init().paginate(session=session, tree=True)
    return Schemas(data=SchemasFunctionPaginateItem(**db_model_list))


@router.get('/{}.params'.format(name), name="get {}".format(name))
async def params_models(session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.list" % name])):
    """
    :param session:
    :param auth:
    :return:
    """
    return Schemas(data={})


@router.get('/{}.menus'.format(name), name="get {}".format(name))
async def menus_permission_models(session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user)):
    """
    :param session:
    :param auth:
    :return:
    """

    def query_fun(nodes):
        permissionClass = CrudFunctionPermission.model_class
        return nodes.filter(getattr(permissionClass, 'is_menu').is_(True)).filter(getattr(permissionClass, 'scope').in_(auth.scopes))

    db_model_list = CrudFunctionPermission.init().get_tree(session=session, query_function=query_fun)
    return SchemasFunctionMenuStatusResponse(data=db_model_list)


@router.post('/{}.menus.post'.format(name), name="post {}".format(name))
async def menus_permission_models(session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user)):
    """
    :param session:
    :param auth:
    :return:
    """

    def query_fun(nodes):
        return nodes.filter(getattr(CrudFunctionPermission.init().model_class, 'is_menu').is_(True)).filter(
            getattr(CrudFunctionPermission.init().model_class, 'scope').in_(auth.scopes))

    db_model_list = CrudFunctionPermission.init().get_tree(session=session, query_function=query_fun)
    return SchemasFunctionMenuStatusResponse(data=db_model_list)


@router.get('/{}/{{pk}}'.format(name), name="get {}".format(name))
async def get_model(pk: int, session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.get" % name])):
    """
    :param pk:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionPermission.init(session=session).first(pk=pk)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    return Schemas(data=SchemasFunctionResponse(**db_model.to_dict()))


@router.post('/{}'.format(name), name="get {}".format(name))
async def store_model(item: SchemasFunctionStoreUpdate, session: Session = Depends(_session),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.store" % name])):
    """
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionPermission.init(session=session).first(where=("name", item.name))
    if db_model is not None:
        return SchemasError(message="数据已经存在！")
    bool_model = CrudFunctionPermission.init(session=session).store(item=item)
    return Schemas(data=SchemasFunctionResponse(**bool_model.to_dict()))


@router.put("/{}/{{pk}}".format(name), name="update {}".format(name))
async def update_put_model(pk: int, item: SchemasFunctionStoreUpdate, session: Session = Depends(_session),
                           auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.update" % name])):
    """
    :param pk:
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionPermission.init(session=session).first(pk=pk)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    bool_model = CrudFunctionPermission.init(session=session).update(pk=pk, item=item)
    return Schemas(data=SchemasFunctionResponse(**bool_model.to_dict()))


@router.patch("/{}/{{pk}}".format(name), name="update {}".format(name))
async def update_patch_model(pk: int, item: SchemasFunctionStoreUpdate, session: Session = Depends(_session),
                             auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.update" % name])):
    """
    :param pk:
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionPermission.init(session=session).first(pk=pk)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    bool_model = CrudFunctionPermission.init(session=session).update(pk=pk, item=item, exclude_unset=True)
    return Schemas(data=SchemasFunctionResponse(**bool_model.to_dict()))


@router.patch("/{}/{{pk}}/move_inside".format(name), name="update {}".format(name))
async def move_inside_permission_model(pk: int, inside_id: int, session: Session = Depends(_session),
                                       auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.update" % name])):
    """
    移动到 inside_id 下
    :param pk:
    :param inside_id:
    :param session:
    :param auth:
    :return:
    """
    bool_model = CrudFunctionPermission.init(session=session).move_inside(inside_id=inside_id, pk=pk)
    return Schemas(data=bool_model)


@router.patch("/{}/{{pk}}/move_after".format(name), name="update {}".format(name))
async def move_after_permission_model(pk: int, after_id: int, session: Session = Depends(_session),
                                      auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.update" % name])):
    """
    移动到 after_id 后
    :param pk:
    :param after_id:
    :param session:
    :param auth:
    :return:
    """
    bool_model = CrudFunctionPermission.init(session=session).move_after(after_id=after_id, pk=pk)
    return Schemas(data=bool_model)


@router.delete("/{}/{{pk}}".format(name), name="delete {}".format(name))
async def delete_model(pk: int, session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.delete" % name])):
    """
    :param pk:
    :param session:
    :param auth:
    :return:
    """
    bool_model = CrudFunctionPermission.init(session=session).delete(pk=pk)
    return Schemas(data=bool_model)


@router.delete("/{}".format(name), name="deletes {}".format(name))
async def delete_models(pks: List[int], session: Session = Depends(_session),
                        auth: SchemasFunctionScopes = Security(auth_user, scopes=permission_scopes + ["%s.delete" % name])):
    """
    :param pks:
    :param session:
    :param auth:
    :return:
    """
    bool_model = CrudFunctionPermission.init(session=session).delete(pks=pks)
    return Schemas(data=bool_model)



def queue_function_data(module, function_name, function_data=None, auth=None):
    """
    队列方法 需要的参数
    :param module:
    :param function_name:
    :param function_data:
    :param auth:
    :return:
    """
    if auth and type(function_data) is dict:
        function_data.update({"auth": {"prefix": auth.prefix, "username": auth.user.username}})
    return {
        "module": module,
        "type": "function",
        "function_name": function_name,
        "function_data": function_data if function_data else {}
    }



from datetime import datetime
from typing import Optional, List
from pydantic import BaseModel

from yao.schema import SchemaPrefixNames, SchemasPaginate
from yao.function.permission.schema import SchemasFunctionMiniResponse, SchemasFunctionMenuMiniResponse


class SchemasFunctionAppointmentResponse(SchemaPrefixNames):
    """角色 返回"""
    uuid: Optional[str] = None
    name: Optional[str] = None
    permissions: Optional[List[SchemasFunctionMiniResponse]] = None
    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True

class SchemasFunctionMiniAppointmentResponse(SchemaPrefixNames):
    """角色 返回"""
    uuid: Optional[str] = None
    name: Optional[str] = None

    class Config:
        orm_mode = True

class SchemasFunctionAppointmentPaginateItem(SchemasPaginate):
    items: List[SchemasFunctionAppointmentResponse]


class SchemasFunctionAppointmentStoreUpdate(BaseModel):
    """授权角色 提交"""
    prefix: Optional[str] = None
    name: Optional[str] = None
    scopes: Optional[str] = None
    permissions: Optional[List[str]] = None
    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态


class SchemasParams(BaseModel):
    permissions: Optional[List[SchemasFunctionMenuMiniResponse]] = None



from yao.crud import Operation
from yao.function.model import ModelFunctionAppointments, ModelFunctionPermissions
from yao.function.appointment.schema import SchemasFunctionAppointmentStoreUpdate


class CrudFunctionAppointment(Operation):
    """用户表操作"""
    model_class = ModelFunctionAppointments
    relationships = {
        "permissions": ModelFunctionPermissions
    }

    def store(self, session, item: SchemasFunctionAppointmentStoreUpdate = None, **kwargs):
        if item.permissions:
            _relation = Operation.init(model_class=ModelFunctionPermissions).get(session=session, where=("uuid", 'in_', item.permissions))
            item.scopes = " ".join([relation.scope for relation in _relation])
        return super().store(session=session, item=item, **kwargs)

    def update(self, session, item: SchemasFunctionAppointmentStoreUpdate = None, **kwargs):
        if item.permissions:
            _relation = Operation.init(model_class=ModelFunctionPermissions).get(session=session, where=("uuid", 'in_', item.permissions))
            item.scopes = " ".join([relation.scope for relation in _relation])
        return super().update(session=session, item=item, **kwargs)



import os
import json
import requests
from datetime import datetime
from urllib.parse import quote

base_attempts = 3


class MP(object):
    app_id = None
    app_secret = None
    cache_file = None
    __errors = []

    def __init__(self, app_id=None, app_secret=None, cache_file=None):
        self.app_id = app_id
        self.app_secret = app_secret
        self.__errors = []
        if cache_file:
            self.cache_file = cache_file
        else:
            try:
                from config import MP_CACHE_FILE
            except:
                MP_CACHE_FILE = os.path.join(os.path.dirname(__file__), "mp.token.%s" % app_id)
            self.cache_file = MP_CACHE_FILE

    def message(self, openid=None, template_id=None, url=None, miniprogram=None, client_msg_id=None, data={}, attempts=0):
        """
        发送订阅信息
        Args:
            openid:
            template_id:
            url:
            miniprogram:
            client_msg_id:
            data:
            attempts:

        Returns:

        """
        res = requests.post("https://api.weixin.qq.com/cgi-bin/message/template/send", params={
            "access_token": self.get_token()
        }, json={
            "touser": openid,
            "template_id": template_id,
            "url": url,
            "miniprogram": miniprogram,
            "client_msg_id": client_msg_id,
            "data": data
        }
                            )
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.message(openid=openid, template_id=template_id, url=url, miniprogram=miniprogram, client_msg_id=client_msg_id, data=data, attempts=attempts + 1)
            else:
                return {}

    def code_to_auto_info(self, code, attempts=0):
        """
        code 换取授权信息
        Args:
            code:
            attempts:

        Returns:

        """
        res = requests.get("https://api.weixin.qq.com/sns/oauth2/access_token?appid=APPID&secret=SECRET&code=CODE&grant_type=authorization_code", params={
            "grant_type": "authorization_code", "appid": self.app_id, "secret": self.app_secret, "code": code
        })
        if res.status_code == 200:
            data = res.json()
            return self.__get_auth_info(access_token=data.get("access_token"), openid=data.get("openid"))
        else:
            if attempts < base_attempts:
                return self.code_to_auto_info(code=code, attempts=attempts + 1)
            else:
                return {}

    def __get_auth_info(self, access_token, openid, attempts=0):
        """
        获取用户信息
        Args:
            access_token:
            openid:
            attempts:

        Returns:

        """
        res = requests.get("https://api.weixin.qq.com/sns/userinfo", params={
            "access_token": access_token, "openid": openid, "lang": "zh_CN"
        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.__get_auth_info(access_token=access_token, openid=openid, attempts=attempts + 1)
            else:
                return {}

    def code_to_openid(self, code, attempts=0):
        """
        code 换取授权信息
        Args:
            code:
            attempts:

        Returns:

        """
        res = requests.get("https://api.weixin.qq.com/sns/oauth2/access_token", params={
            "grant_type": "authorization_code", "appid": self.app_id, "secret": self.app_secret, "code": code
        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.code_to_openid(code=code, attempts=attempts + 1)
            else:
                return {}

    def get_or_to_auth_url(self, redirect_uri, scope="snsapi_base", state="L", fastapi_return=False):
        """
        获取或者到授权地址
        Args:
            redirect_uri:
            scope: snsapi_base | snsapi_userinfo
            state:
            fastapi_return:
        Returns:

        """
        redirect_uri = quote(redirect_uri)
        url = "https://open.weixin.qq.com/connect/oauth2/authorize?appid=%s&redirect_uri=%s&response_type=code&scope=%s&state=%s#wechat_redirect" % (
            self.app_id, redirect_uri, scope, state)
        if fastapi_return:
            from fastapi.responses import RedirectResponse
            return RedirectResponse(url=url)
        return url

    def get_token(self):
        cache_content = self.__get_file_content_or_create_file(self.cache_file)
        if "access_token" in cache_content and "expires_time" in cache_content and cache_content.get("expires_time") > datetime.now().timestamp():
            return cache_content.get("access_token")
        else:
            token = self.__get_token()
            token.update({
                "expires_time": datetime.now().timestamp() + token.get("expires_in", 0)
            })
            self.__create_file(self.cache_file, content=json.dumps(token), mode="w")
            return token.get("access_token")

    def __get_token(self, attempts=0):
        res = requests.get("https://api.weixin.qq.com/cgi-bin/token", params={
            "grant_type": "client_credential", "appid": self.app_id, "secret": self.app_secret
        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.__get_token(attempts=attempts + 1)
            else:
                return {}

    def __create_file(self, path, content=None, mode="wb", **kwargs):
        path_dir = os.path.dirname(path)
        if not os.path.exists(path_dir):
            os.makedirs(path_dir)
        if not os.path.isdir(path):
            with open(path, mode, **kwargs) as f:
                f.write(content)
            return True
        return False

    def __get_file_content(self, path, mode="rb", **kwargs):
        """获取文件内容"""
        if os.path.isfile(path):
            with open(path, mode, **kwargs) as f:
                return f.read()
        return bytes()

    def __get_file_content_or_create_file(self, path, read_mode="r", write_mode="w"):
        if not os.path.exists(self.cache_file):
            self.__create_file(path, content="{}", mode=write_mode)
            return dict()
        else:
            bytes_content = self.__get_file_content(path, mode=read_mode)
            return json.loads(bytes_content)

#
# if __name__ == '__main__':
#     from config import MPAPPID, MPAPPSECRET
#
#     mp = MP(app_id=MPAPPID, app_secret=MPAPPSECRET)
#     print(mp.message(openid="o9vB96AmVEZCLO09jPaplvvRDlok", template_id="8BNzzuBzXd_ttcr2KHAbDEuJLyQ5IVKXYR1YiRLJ2Sk", data={
#         "thing2": {
#             "value": "公司高抛"
#         },
#         "thing3": {
#             "value": "软件园43号楼"
#         },
#         "time5": {
#             "value": "2022年12月27日 13:48:43"
#         },
#     }))



from sqlalchemy import Column, String, Boolean, Text, Integer, ForeignKey, event, Table, JSON, TIMESTAMP
from sqlalchemy.orm import relationship, declared_attr
from sqlalchemy_mptt.mixins import BaseNestedSets

from yao.method import plural
from yao.db import Model, Engine

name = __name__[:__name__.rfind(".")].capitalize()

"""职能信息"""
function_name = plural("%s" % name)

"""公司信息"""
function_company_name = plural("%s.company" % name)
function_company_table_name = function_company_name.replace('.', '_')

"""公司部门"""
function_department_name = plural("%s.department" % name)
function_department_table_name = function_department_name.replace('.', '_')

"""员工职位"""
function_appointment_name = plural("%s.appointment" % name)
function_appointment_table_name = function_appointment_name.replace('.', '_')

"""权限信息"""
function_permission_name = plural("%s.permission" % name)
function_permission_table_name = function_permission_name.replace('.', '_')

"""后台账号"""
function_user_name = plural("%s.user" % name)
function_user_table_name = function_user_name.replace('.', '_')

"""附件"""
function_annex_name = plural("%s.annex" % name)
function_annex_table_name = function_annex_name.replace('.', '_')

"""日志"""
function_log_name = plural("%s.log" % name)
function_log_table_name = function_log_name.replace('.', '_')

"""队列"""
function_queue_name = plural("%s.queue" % name)
function_queue_table_name = function_queue_name.replace('.', '_')

SYSTEM_PERMISSIONS = [
    {
        "name": "职能信息",
        "scope": function_name,
        "icon": "ElementPlus",
        "children": [
            {"name": "公司信息", "scope": function_company_name, "icon": "Refrigerator"},
            {"name": "公司部门", "scope": function_department_name, "icon": "Help"},
            {"name": "员工职位", "scope": function_appointment_name, "icon": "User"},
            {"name": "权限管理", "scope": function_permission_name, "icon": "KnifeFork", "is_menu": False,
             "action": [{"name": "菜单", "scope": "menus"}, {"name": "树", "scope": "tree"}]},
            {"name": "后台账号", "scope": function_user_name, "icon": "User"},
            {"name": "附件文件", "scope": function_annex_name, "icon": "Folder", "is_menu": False},
            {"name": "操作日志", "scope": function_log_name, "icon": "Document"},
            {"name": "后台队列", "scope": function_queue_name, "icon": "Finished"},
        ]
    }
]


class BaseModel(Model):
    __abstract__ = True
    remarks = Column(String(150), nullable=True, comment="备注")
    sort = Column(Integer, default=999, comment="排序")
    status = Column(Boolean, default=True, comment="状态")


class ModelFunctionCompanies(BaseModel):
    """ 公司信息 """
    __tablename__ = function_company_table_name
    name = Column(String(30), nullable=False, unique=True, comment="公司名称")
    prefix_name = Column(String(20), nullable=False, unique=True, comment="公司前缀")
    contact_name = Column(String(20), nullable=True, comment="联系人")
    contact_phone = Column(String(20), nullable=True, comment="联系电话")


class BaseCompanyModel(BaseModel):
    __abstract__ = True

    @declared_attr
    def prefix(self):
        return Column(String(20), ForeignKey('%s.prefix_name' % function_company_table_name, ondelete="CASCADE", onupdate="CASCADE"), index=True, nullable=False,
                      comment="公司前缀")


class ModelFunctionDepartments(BaseCompanyModel, BaseNestedSets):
    """公司部门"""
    __tablename__ = function_department_table_name
    name = Column(String(40), unique=True, nullable=False, comment="名称")


class ModelFunctionPermissions(BaseModel, BaseNestedSets):
    """ 权限 """
    __tablename__ = function_permission_table_name
    sqlalchemy_mptt_pk_name = "uuid"
    name = Column(String(15), unique=True, nullable=False, comment="名称")
    icon = Column(String(20), nullable=True, comment="ICO")
    scope = Column(String(50), nullable=False, unique=True, comment="Scope")
    path = Column(String(50), nullable=False, comment="Path")
    is_menu = Column(Boolean, default=True, comment="是否为菜单")
    is_action = Column(Boolean, default=True, comment="是否为动作权限")


@event.listens_for(ModelFunctionPermissions.scope, 'set')
def permission_receive_before_insert(target, value: str, old_value, initiator):
    target.path = "/%s" % value.replace('.', '/').lower()


@event.listens_for(ModelFunctionPermissions.scope, 'modified')
def permission_receive_before_insert(target, initiator):
    target.path = "/%s" % target.scope.replace('.', '/').lower()


class ModelFunctionAppointments(BaseCompanyModel):
    """员工职位"""
    __tablename__ = function_appointment_table_name
    department = Column(String(40), ForeignKey('%s.name' % function_department_table_name, ondelete="CASCADE", onupdate="CASCADE"), index=True, nullable=True, comment="公司部门")
    name = Column(String(40), unique=True, nullable=False, comment="名称")
    scopes = Column(Text, nullable=False, comment="Scope")
    #  lazy="joined"
    permissions = relationship(ModelFunctionPermissions, backref='appointments', secondary=Table(
        "%s_appointment_has_permissions" % function_name.replace('.', '_'),
        Model.metadata,
        Column('per_id', Integer, ForeignKey("%s.id" % function_permission_table_name, ondelete="CASCADE"), primary_key=True, comment="权限"),
        Column('app_id', Integer, ForeignKey("%s.id" % function_appointment_table_name, ondelete="CASCADE"), primary_key=True, comment="职位")
    ))


class ModelFunctionUsers(BaseCompanyModel, BaseNestedSets):
    """登录用户"""
    __tablename__ = function_user_table_name
    sqlalchemy_mptt_pk_name = "uuid"
    username = Column(String(40), nullable=False, unique=True, comment="名称")
    password = Column(String(128), nullable=False, comment="密码")
    user_phone = Column(String(11), nullable=True, comment="手机")
    available = Column(Boolean, default=1, comment="是否有效")
    children_ids = Column(JSON, nullable=True, comment="子孙ids")
    auth_data = Column(JSON, nullable=True, comment="授权数据")

    permissions = relationship(ModelFunctionPermissions, backref='function_users', secondary=Table(
        "%s_user_has_permissions" % function_name.replace('.', '_'),
        Model.metadata,
        Column('per_id', Integer, ForeignKey("%s.id" % function_permission_table_name, ondelete="CASCADE"), primary_key=True, comment="权限"),
        Column('use_id', Integer, ForeignKey("%s.id" % function_user_table_name, ondelete="CASCADE"), primary_key=True, comment="用户"),
    ))
    appointments = relationship(ModelFunctionAppointments, backref='function_users', secondary=Table(
        "%s_user_has_appointments" % function_name.replace('.', '_'),
        Model.metadata,
        Column('app_id', Integer, ForeignKey("%s.id" % function_appointment_table_name, ondelete="CASCADE"), primary_key=True, comment="职位"),
        Column('use_id', Integer, ForeignKey("%s.id" % function_user_table_name, ondelete="CASCADE"), primary_key=True, comment="用户"),
    ))


class ModelFunctionAnnexes(BaseCompanyModel):
    """ 附件 """
    __tablename__ = function_annex_table_name

    filename = Column(String(50), nullable=False, comment="文件名")
    content_type = Column(String(100), nullable=False, comment="类型")
    path = Column(String(100), nullable=True, comment="路径")
    md5 = Column(String(32), nullable=True, comment="md5", index=True)
    size = Column(Integer, nullable=False, comment="SIZE")
    width = Column(String(100), nullable=True, comment="宽")
    height = Column(String(190), nullable=True, comment="高")

    @property
    def preview_path(self):
        import os
        from config import STATIC_URL
        return os.path.join(STATIC_URL, self.path)


class ModelFunctionLogs(BaseCompanyModel):
    """ 日志 """
    __tablename__ = function_log_table_name
    scope = Column(String(100), nullable=True, comment="scope")
    methods = Column(String(32), nullable=True, comment="methods", index=True)
    data = Column(JSON, nullable=True, comment="Data")
    username = Column(String(32), ForeignKey("%s.username" % function_user_table_name, onupdate="CASCADE", ondelete="CASCADE"), index=True, nullable=True, comment="用户")


class ModelFunctionQueues(BaseCompanyModel):
    """队列"""
    __tablename__ = function_queue_table_name
    username = Column(String(32), ForeignKey("%s.username" % function_user_table_name, onupdate="CASCADE", ondelete="CASCADE"), index=True, nullable=True, comment="用户")
    priority = Column(Integer, nullable=True, comment="优先级 小的最高")
    scope = Column(String(100), nullable=True, comment="scope")
    data = Column(JSON, nullable=True, comment="Data")
    key = Column(Text, nullable=True, comment="去重Key")
    start_at = Column(TIMESTAMP, nullable=True, comment="开始时间")
    stop_at = Column(TIMESTAMP, nullable=True, comment="结束时间")
    progress = Column(String(8), nullable=True, comment="进度条")
    progress_text = Column(String(100), nullable=True, comment="进度条")
    retry = Column(Integer, nullable=True, comment="重试次数")
    queue_status = Column(Integer, nullable=True, comment="状态 0未运行 1成功 2失败")



from abc import ABCMeta
from typing import Union, List, Tuple, Optional

from sqlalchemy import or_
from sqlalchemy.sql.expression import func
from pydantic import BaseModel


class Operation(object, metaclass=ABCMeta):
    """
    模型类
    """
    model_class = None
    model_pk = "id"
    relationships: dict = {}  # 多对多时使用
    relationship_pk: str = "uuid"  # 多对多传来的值对应的字段　　　
    relations: dict = {}  # 关联表 用于join等
    instance = None
    query = None

    limit = None
    page = None
    order = None

    def __init__(self, model_class=None, model_pk=None):
        """
        实例化对象
        :param model_class:
        :param model_pk:
        """
        if model_class:
            """设置模型类"""
            self.model_class = model_class
        if model_pk:
            """设置主銉"""
            self.model_pk = model_pk

    @classmethod
    def init(cls, cache=False, **kwargs):
        """
        初始化实例对象
        :param cache:
        :param kwargs:
        :return:
        """
        if not cache or not cls.instance:
            cls.instance = cls(**kwargs)
        return cls.instance

    def __init_query(self, session, field=None):
        """
        初始化查询
        :param field:
        :return:
        """
        self.limit = None
        self.page = None
        self.order = None
        self.query = session.query(field if field else self.model_class)
        return self.query

    def __init_params(self, **kwargs):
        """
        处理kwargs 参数到 params
        :param kwargs:
        :return:
        """
        [(getattr(self, "_%s" % params)(params_value) if hasattr(self, "_%s" % params) else None) for params, params_value in kwargs.items()]
        return self

    def _screen_params(self, params: BaseModel):
        """
        处理screen_params 参数到 params
        :param params:
        :return:
        """
        self.__init_params(**params.dict())
        return self

    def __w_tool(self, model_class, where):
        if where:
            if len(where) == 2:
                return getattr(model_class, where[0]) == where[1]
            elif len(where) == 3:
                if where[1] in ["in"] and type(where[2]) in [list, tuple]:
                    return getattr(getattr(model_class, where[0]), "in_")(where[2])
        return None

    def __filter_query(self, where: Union[list, tuple, None] = None, query=None, model=None):
        """
        过滤数据条件
        :param where:
        :param query:
        :param model:
        :return:
        """
        _query = query if query else self.query
        model_class = model if model else self.model_class
        if bool(where) and (type(where) == tuple or type(where) == list):
            if len(where) == 2:
                """('content', '西')"""
                if where[0][:2] != "__":
                    _query = _query.filter(getattr(model_class, where[0]) == where[1])
                else:
                    """ ("__or", [("a", b), ("c", "in", [1,2])]) """
                    _filters = [self.__w_tool(model_class, fil) for fil in where[1]]
                    if where[0][2:] != "or":
                        _query = _query.filter(or_(*[f for f in _filters if f is not None]))
            elif len(where) == 3:
                if where[1] == "==" or where[1] == "=" or where[1] == "eq":
                    """('content', '==', '西')"""
                    if where[2] != "_#None":
                        _query = _query.filter(getattr(model_class, where[0]) == where[2])
                    else:
                        _query = _query.filter(or_(getattr(model_class, where[0]) == None, getattr(model_class, where[0]) == ""))
                elif where[1] == "!=" or where[1] == "<>" or where[1] == "><" or where[1] == "neq" or where[1] == "ne":
                    """('content', '!=', '西')"""
                    if where[2] != "_#None":
                        _query = _query.filter(getattr(model_class, where[0]) != where[2])
                    else:
                        _query = _query.filter(or_(getattr(model_class, where[0]) != None, getattr(model_class, where[0]) != ""))
                elif where[1] == ">" or where[1] == "gt":
                    """('content', '>', '西')"""
                    _query = _query.filter(getattr(model_class, where[0]) > where[2])
                elif where[1] == ">=" or where[1] == "ge":
                    """('content', '>=', '西')"""
                    _query = _query.filter(getattr(model_class, where[0]) >= where[2])
                elif where[1] == "<" or where[1] == "lt":
                    """('content', '<', '西')"""
                    _query = _query.filter(getattr(model_class, where[0]) < where[2])
                elif where[1] == "<=" or where[1] == "le":
                    """('content', '<=', '西')"""
                    _query = _query.filter(getattr(model_class, where[0]) <= where[2])
                elif where[1] in ["like", "ilike"]:
                    if not where[2] is None:
                        """('content', 'like', '西')"""
                        _query = _query.filter(
                            getattr(getattr(model_class, where[0]), where[1])("%" + where[2] + "%")
                        )
                elif where[1] in ["or"]:
                    if type(where[0]) == tuple or type(where[0]) == list:
                        """(['name','content'], 'or', '西')"""
                        _filters = [(getattr(model_class, fil) == where[2]) for fil in where[0]]
                        _query = _query.filter(or_(*_filters))

                    if type(where[0]) == str and (type(where[2]) == tuple or type(where[2]) == list):
                        """('name', 'or', ['西', '西'])"""
                        _filters = [(getattr(model_class, where[0]) == fil) for fil in where[2]]
                        _query = _query.filter(or_(*_filters))
                elif where[1] in ["or_like", "or_ilike"]:
                    if type(where[0]) == tuple or type(where[0]) == list:
                        """(['name','content'], 'or_like', '西')"""
                        _filters = [(getattr(getattr(model_class, fil), where[1][3:])("%" + where[2] + "%")) for
                                    fil in where[0]]
                        _query = _query.filter(or_(*_filters))
                    if type(where[0]) == str and (type(where[2]) == tuple or type(where[2]) == list):
                        """('name', 'or_like', ['西', '西'])"""
                        _filters = [(getattr(getattr(model_class, where[0]), where[1][3:])("%" + fil + "%")) for
                                    fil in where[2]]
                        _query = _query.filter(or_(*_filters))
                elif where[1] in ["between"] and type(where[2]) in [list, tuple] and len(where[2]) == 2:
                    _query = _query.filter(getattr(getattr(model_class, where[0]), where[1])(where[2][0], where[2][1]))
                elif where[1] in ["datebetween"] and type(where[2]) in [list, tuple] and len(where[2]) == 2:
                    _query = _query.filter(getattr(getattr(model_class, where[0]), 'between')(where[2][0], where[2][1]))
                elif where[1] in ["datetimebetween"] and type(where[2]) in [list, tuple] and len(where[2]) == 2:
                    _query = _query.filter(getattr(getattr(model_class, where[0]), 'between')("%s" % where[2][0], "%s 23:59:59" % where[2][1]))
                elif where[1] in ["in"] and type(where[2]) in [list, tuple]:
                    _query = _query.filter(getattr(getattr(model_class, where[0]), "in_")(where[2]))
                elif where[1] in ["is"]:
                    _query = _query.filter(getattr(getattr(model_class, where[0]), "is_")(where[2]))
                elif where[1] in ["notin"] and type(where[2]) in [list, tuple]:
                    _query = _query.filter(getattr(getattr(model_class, where[0]), "notin_")(where[2]))
                else:
                    _query = _query.filter(getattr(getattr(model_class, where[0]), where[1])(where[2]))
        if not query:
            self.query = _query
        return _query

    def _where(self, where: Optional[Union[List[tuple], List[list], Tuple[tuple], Tuple[list], list, tuple, str]] = None, *args):
        """
        条件集合
        :param where:
        :param args:
        :return:
        """

        if where:
            if (type(where) == list or type(where) == tuple) and type(where[0]) == str:
                where = [where]
            elif type(where) == str:
                where = [(where, *args)]
            [self.__filter_query(where=w) for w in where]

    def _pk(self, pk: int):
        """
        设定查询 主键
        :param pk:
        :return:
        """
        self._where((self.model_pk, pk))

    def _uuid(self, uuid: Union[int, str]):
        """
        设定查询 uuid
        :param uuid:
        :return:
        """
        self._where(("uuid", uuid))

    def _uuids(self, uuids: List[Union[int, str]]):
        """
        设定查询 uuid
        :param uuids:
        :return:
        """
        self._where(("uuid", "in", uuids))

    def _limit(self, limit):
        """
        设置 limit
        :param limit:
        :return:
        """
        self.limit = limit

    def _page(self, page):
        """
        设置 page
        :param page:
        :return:
        """
        self.page = page

    def _order(self, order):
        """
        设置 order
        :param order:
        :return:
        """
        self.order = order

    def _rand(self, limit):
        """
        设置 order
        :param order:
        :return:
        """
        self.limit = limit
        self.query = self.query.order_by(func.rand())

    def _query_limit_page(self):
        if self.limit:
            self.query = self.query.limit(self.limit)
            if self.page:
                self.query = self.query.offset((self.page - 1) * self.limit)

    def _query_order(self):
        if bool(self.order):
            if type(self.order) == tuple or type(self.order) == list:  # 设置排序
                for attr_item in self.order:
                    self.query = self.query.order_by(getattr(getattr(self.model_class, attr_item[0]), attr_item[1])())

    def get(self, session, field=None, close: bool = False, **kwargs):
        """
        获取列表
        :param session:
        :param field:
        :param close:
        :param kwargs:
        :return:
        """
        self.__init_query(session=session, field=field)
        self.__init_params(**kwargs)
        self._query_order()
        self._query_limit_page()
        response = self.query.all()
        close and session.close()
        return response

    def count(self, session, close: bool = False, **kwargs):
        """
        获取列表
        :param session:
        :param close:
        :param kwargs:
        :return:
        """
        self.__init_query(session=session)
        self.__init_params(**kwargs)
        response = self.query.count()
        close and session.close()
        return response

    def paginate(self, session, field=None, limit=None, page=None, order=None, close: bool = False, tree: bool = False, **kwargs):
        """
        分页 操作
        :param session:
        :param field:
        :param limit:
        :param page:
        :param order:
        :param close:
        :param tree:
        :param kwargs:
        :return:
        """
        import math
        total = self.count(session=session, **kwargs)
        if tree:
            items = self.get_tree(session=session, json=False)
            items = [item.get("node") for item in items]
        else:
            items = self.get(session=session, field=field, limit=limit, page=page, order=order, close=close, **kwargs)
        pages = math.ceil(total / self.limit) if type(total) is int and type(self.limit) is int and self.limit != 0 else 1
        return {
            "items": items,  # 当前页的数据列表
            "pages": pages,  # 总页数
            "total": total,  # 总条数
            "page": self.page,  # 当前页
            "limit": self.limit,  # 页条数
        }

    def first(self, session, field=None, close: bool = False, **kwargs):
        """
        获取条件第一个
        :param session:
        :param field:
        :param close:
        :param kwargs:
        :return:
        """
        self.__init_query(session=session, field=field)
        self.__init_params(**kwargs)
        self._query_order()
        response = self.query.first()
        close and session.close()
        return response

    def update(self, session, where=None, item: BaseModel = None, data: dict = None, commit: bool = True, refresh: bool = True, close: bool = False, exclude_unset=True,
               event: bool = False, synchronize_session="evaluate", **kwargs):
        """
        更新数据
        :param session:
        :param where:
        :param item:
        :param data:
        :param commit:
        :param refresh:
        :param close:
        :param exclude_unset:
        :param event:
        :param synchronize_session:
        :param kwargs:
        :return:
        """
        if event:
            obj_item = self.first(session=session, where=where, **kwargs)
            # relationships
            if bool(self.relationships):
                for (relation, relation_class) in self.relationships.items():
                    if hasattr(item, relation) and getattr(item, relation):
                        relation__ = Operation.init(model_class=relation_class)
                        _relation = relation__.get(session=session, where=(relation__.relationship_pk, 'in_', getattr(item, relation)))
                        setattr(item, relation, _relation)
            # relationships
            [setattr(obj_item, k, val) for k, val in item.dict(exclude_unset=exclude_unset).items()]
            commit and session.commit()
            close and session.close()
        else:
            # relationships
            obj_item = None
            if bool(self.relationships):
                for (relation, relation_class) in self.relationships.items():
                    obj_item = obj_item or self.first(session=session, where=where)
                    if hasattr(item, relation) and bool(getattr(item, relation)):

                        relation__ = Operation.init(model_class=relation_class)
                        _relation = relation__.get(session=session, where=(relation__.relationship_pk, 'in_', getattr(item, relation)))
                        setattr(obj_item, relation, _relation)
                    elif not bool(getattr(item, relation)) and relation in item.dict(exclude_unset=exclude_unset):
                        [getattr(obj_item, relation).remove(rela) for rela in list(getattr(obj_item, relation))]  # 出错
                    hasattr(item, relation) and delattr(item, relation)
            # relationships
            self.__init_query(session=session)
            self.__init_params(where=where, **kwargs)
            self.query.update(item.dict(exclude_unset=exclude_unset), synchronize_session=synchronize_session)
            commit and session.commit()
            close and session.close()

    def store(self, session, item: BaseModel = None, data: dict = None, commit: bool = True, refresh: bool = True, close: bool = False, **kwargs):
        """
        创建模型数据
        :param session:
        :param item:
        :param data:
        :param commit:
        :param refresh:
        :param close:
        :param kwargs:
        :return:
        """

        # relationships
        if bool(self.relationships):
            for (relation, relation_class) in self.relationships.items():
                if hasattr(item, relation) and bool(getattr(item, relation)):
                    relation__ = Operation.init(model_class=relation_class)
                    _relation = relation__.get(session=session, where=(relation__.relationship_pk, 'in_', getattr(item, relation)))
                    setattr(item, relation, _relation)
        # relationships

        # 创建实例
        db_item = self.model_class(**item.dict(exclude_unset=True))
        session.add(db_item)
        commit and session.commit()
        refresh and session.refresh(db_item)
        close and session.close()
        return db_item

    def many_store(self, session, items: List[BaseModel], commit: bool = True, close: bool = False, **kwargs):
        """
        批量创建模型数据
        :param session:
        :param items:
        :param commit:
        :param close:
        :param kwargs:
        :return:
        """
        db_items = [self.model_class(**item.dict(exclude_unset=True)) for item in items]
        session.add_all(db_items)
        commit and session.commit()
        close and session.close()
        return db_items

    def delete(self, session, commit: bool = True, close: bool = False, event: bool = False, **kwargs):
        """
        删除多模型数据
        :param session:
        :param commit:
        :param close:
        :param event:
        :param kwargs:
        :return:
        """
        self.__init_query(session=session)
        self.__init_params(**kwargs)
        self._query_order()

        response = [session.delete(u) for u in self.query.all()] if event else self.query.delete()

        commit and session.commit()
        close and session.close()
        return response

    def update_or_store_model(self, session, where=None, item: BaseModel = None, store_item: BaseModel = None, data: dict = None, commit: bool = True, refresh: bool = True,
                              close: bool = False, exclude_unset: bool = True, event: bool = False, update_event: bool = False, **kwargs):
        """
        更新或者创建
        :param session:
        :param where:
        :param item:
        :param store_item:
        :param data:
        :param commit:
        :param refresh:
        :param close:
        :param exclude_unset:
        :param event:
        :param update_event:
        :param kwargs:
        :return:
        """
        instance = self.first(session=session, where=where, **kwargs)
        if instance:
            return self.update(session=session, item=item, data=data, pk=getattr(instance, self.model_pk), exclude_unset=exclude_unset, commit=commit, refresh=refresh, close=close,
                               event=update_event, **kwargs)
        else:
            return self.store(session=session, item=store_item if store_item else item, data=data, commit=commit, refresh=refresh, close=close, event=event, **kwargs)

    def find_or_store_model(self, session, where=None, item: BaseModel = None, data: dict = None, commit: bool = True, refresh: bool = True, close: bool = False, **kwargs):
        """
        查找或者创建
        :param session:
        :param where:
        :param item:
        :param data:
        :param commit:
        :param refresh:
        :param close:
        :param kwargs:
        :return:
        """
        instance = self.first(session=session, where=where, **kwargs)
        if not instance:
            return self.store(session=session, item=item, data=data, commit=commit, refresh=refresh, close=close, **kwargs)
        return instance

    def get_tree(self, session, json=True, json_fields=None, query_function=None, where: Union[list, tuple] = None, **kwargs):
        """
        获取树
        :param session:
        :param json:
        :param json_fields:
        :param query_function:
        :param where:
        :param kwargs:
        :return:
        """

        def query_fun(nodes, _where: Union[list, tuple] = None):
            if _where:
                if (type(_where) == list or type(_where) == tuple) and type(_where[0]) == str:
                    _where = [_where]
                for w in where:
                    nodes = self.__filter_query(where=w, query=nodes)
            return nodes

        return self.model_class.get_tree(session=session, json=json, json_fields=lambda node: node.to_dict() if json is True and not json_fields else json_fields,
                                         query=query_function if query_function else lambda nodes: query_fun(nodes=nodes, _where=where))

    def update_children_uuids(self, **kwargs):
        """
        更新子级uuid
        Args:
            **kwargs:

        Returns:

        """
        objs = self.get(**kwargs)
        for obj in objs:
            ids = []

            def _get_id(obj):
                _ids = [obj.get("node").uuid]
                children = obj.get("children", [])
                if len(children) > 0:
                    for child in children:
                        _ids += _get_id(child)
                return _ids

            for _obj in obj.drilldown_tree():
                ids += _get_id(_obj)
            obj.children_ids = ids
        session = kwargs.get("session")
        commit = kwargs.get("commit", True)
        close = kwargs.get("close", True)
        commit and session.commit()
        close and session.close()
        return True


from time import time


def run_time(func):
    def wrap(*args, **kwargs):
        t1 = time()
        r = func(*args, **kwargs)
        t2 = time()
        print(t2 - t1)
        return r

    return wrap


class LogOperation(Operation):
    @run_time
    def update(self, session, where=None, item: BaseModel = None, data: dict = None, commit: bool = True, refresh: bool = True, close: bool = False, exclude_unset=True,
               event: bool = False, **kwargs):
        return super().update(session=session, where=where, item=item, data=data, commit=commit, refresh=refresh, close=close, exclude_unset=exclude_unset,
                              event=event, **kwargs)



import os
import json
import requests
from datetime import datetime

base_attempts = 3


class AMP(object):
    app_id = None
    app_secret = None
    cache_file = None
    __errors = []

    def __init__(self, app_id=None, app_secret=None, cache_file=None):
        self.app_id = app_id
        self.app_secret = app_secret
        self.__errors = []
        if cache_file:
            self.cache_file = cache_file
        else:
            try:
                from config import AMP_CACHE_FILE
            except:
                AMP_CACHE_FILE = os.path.join(os.path.dirname(__file__), "amp.token.%s" % app_id)
            self.cache_file = AMP_CACHE_FILE

    def code2session(self, js_code, attempts=0):
        """小程序登录"""
        res = requests.get("https://api.weixin.qq.com/sns/jscode2session", params={
            "appid": self.app_id,
            "secret": self.app_secret,
            "js_code": js_code,
            "grant_type": "authorization_code",
        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.code2session(js_code, attempts=attempts + 1)
            else:
                return {}

    def get_wxa_code(self, page=None, scene=None, check_path=False, state="release", width=430, auto_color=False, line_color={"r": 0, "g": 0, "b": 0}, is_hyaline=False,
                     attempts=0):
        """
        获取不限制的小程序码
        Args:
            page: pages/index/index
            scene: a=1
            check_path:
            state:
            width:
            auto_color:
            line_color:
            is_hyaline:
            attempts:

        Returns:

        """
        res = requests.post("https://api.weixin.qq.com/wxa/getwxacodeunlimit", params={
            "access_token": self.get_token()
        }, json={
            "page": page,
            "scene": scene,
            "check_path": check_path,
            "env_version": state,
            "width": width,
            "auto_color": auto_color,
            "line_color": line_color,
            "is_hyaline": is_hyaline
        })
        if res.status_code == 200:
            try:
                return res.json()
            except:
                return res.content
        else:
            if attempts < base_attempts:
                return self.get_wxa_code(page=page, scene=scene, check_path=check_path, state=state, width=width, auto_color=auto_color, line_color=line_color,
                                         is_hyaline=is_hyaline, attempts=attempts + 1)
            else:
                return {}

    def message(self, openid=None, template_id=None, page=None, state="formal", lang="zh_CN", data={}, attempts=0):
        """
        发送订阅信息
        Args:
            openid:
            template_id:
            page:
            state:
            lang:
            data:
            attempts:

        Returns:
        """
        res = requests.post("https://api.weixin.qq.com/cgi-bin/message/subscribe/send", params={
            "access_token": self.get_token()
        }, json={
            "touser": openid,
            "template_id": template_id,
            "page": page,
            "miniprogram_state": state,
            "lang": lang,
            "data": data
        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.message(openid=openid, template_id=template_id, page=page, state=state, lang=lang, data=data, attempts=attempts + 1)
            else:
                return {}

    def custom_message(self, openid=None, content=None, attempts=0):
        res = requests.post("https://api.weixin.qq.com/cgi-bin/message/custom/send", params={
            "access_token": self.get_token()
        }, json={
            "touser": openid,
            "msgtype": "text",
            "text":
                {
                    "content": content
                }
        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.custom_message(openid=openid, content=content, attempts=attempts + 1)
            else:
                return {}

    def uniform_send(self, openid=None, template_id=None, page=None, form_id=None, emphasis_keyword=None, data={}, attempts=0):

        res = requests.post("https://api.weixin.qq.com/cgi-bin/message/wxopen/template/uniform_send", params={
            "access_token": self.get_token()
        }, json={
            "touser": openid,
            "weapp_template_msg": {
                "template_id": template_id,
                "page": page,
                "form_id": form_id,
                "emphasis_keyword": emphasis_keyword,
                "data": data
            }

        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.uniform_send(openid=openid, template_id=template_id, page=page, form_id=form_id, emphasis_keyword=emphasis_keyword, data=data, attempts=attempts + 1)
            else:
                return {}

    def get_token(self):
        cache_content = self.__get_file_content_or_create_file(self.cache_file)
        if "access_token" in cache_content and "expires_time" in cache_content and cache_content.get("expires_time") > datetime.now().timestamp():
            return cache_content.get("access_token")
        else:
            token = self.__get_token()
            token.update({
                "expires_time": datetime.now().timestamp() + token.get("expires_in", 0)
            })
            self.__create_file(self.cache_file, content=json.dumps(token), mode="w")
            return token.get("access_token")

    def get_generate_scheme(self, path=None, query=None, env_version="release", is_expire=True, expire_type=1, expire_interval=1, attempts=0):
        data = {
            "is_expire": is_expire,
            "expire_type": expire_type,
            "expire_interval": expire_interval
        }
        if path:
            data.update({
                "jump_wxa": {
                    "path": path,
                    "query": query,
                    "env_version": env_version
                }
            })
        res = requests.post("https://api.weixin.qq.com/wxa/generatescheme", params={
            "access_token": self.get_token()
        }, json=data)
        if res.status_code == 200:
            res_data = res.json()
            if res_data.get("errcode") == 0:
                return res_data.get("openlink")
            # elif res_data.get("errcode") == 40001:
            #     self.__create_file(path, content="{}", mode="w")
            #     if attempts < base_attempts:
            #         return self.get_generate_scheme(path=path, query=query, env_version=env_version, is_expire=is_expire, expire_type=expire_type, expire_interval=expire_interval,
            #                                         attempts=attempts + 1)
            else:
                self.error("获取scheme码", res_data)
                return False
        else:
            if attempts < base_attempts:
                return self.get_generate_scheme(path=path, query=query, env_version=env_version, is_expire=is_expire, expire_type=expire_type, expire_interval=expire_interval,
                                                attempts=attempts + 1)
            else:
                return False

    def get_generate_url_link(self, path=None, query=None, env_version="release", is_expire=True, expire_type=1, expire_interval=1, cloud_base=None, attempts=0):
        data = {
            "path": path,
            "query": query,
            "is_expire": is_expire,
            "expire_type": expire_type,
            "expire_interval": expire_interval,
            "env_version": env_version
        }

        if cloud_base:
            data.update({
                "cloud_base": cloud_base
            })
        res = requests.post("https://api.weixin.qq.com/wxa/generate_urllink", params={
            "access_token": self.get_token()
        }, json=data)
        if res.status_code == 200:
            res_data = res.json()
            if res_data.get("errcode") == 0:
                return res_data.get("url_link")
            else:
                self.error("获取URLLink", res_data)
                return False
        else:
            if attempts < base_attempts:
                return self.get_generate_url_link(path=path, query=query, env_version=env_version, is_expire=is_expire, expire_type=expire_type, expire_interval=expire_interval,
                                                  cloud_base=cloud_base, attempts=attempts + 1)
            else:
                return False

    def __get_token(self, attempts=0):
        res = requests.get("https://api.weixin.qq.com/cgi-bin/token", params={
            "grant_type": "client_credential", "appid": self.app_id, "secret": self.app_secret
        })
        if res.status_code == 200:
            return res.json()
        else:
            if attempts < base_attempts:
                return self.__get_token(attempts=attempts + 1)
            else:
                return {}

    def __create_file(self, path, content=None, mode="wb", **kwargs):
        path_dir = os.path.dirname(path)
        if not os.path.exists(path_dir):
            os.makedirs(path_dir)
        if not os.path.isdir(path):
            with open(path, mode, **kwargs) as f:
                f.write(content)
            return True
        return False

    def __get_file_content(self, path, mode="rb", **kwargs):
        """获取文件内容"""
        if os.path.isfile(path):
            with open(path, mode, **kwargs) as f:
                return f.read()
        return bytes()

    def __get_file_content_or_create_file(self, path, read_mode="r", write_mode="w"):
        if not os.path.exists(self.cache_file):
            self.__create_file(path, content="{}", mode=write_mode)
            return dict()
        else:
            bytes_content = self.__get_file_content(path, mode=read_mode)
            return json.loads(bytes_content)

    def error(self, text=None, response=None):
        """
        设置或者获取错误
        :param text:
        :param response:
        :return:
        """
        if text is None and response is None:
            return self.__errors
        self.__errors.append({"text": text, "response": response})

#
# if __name__ == '__main__':
#     from config import PROGRAMAPPID, PROGRAMAPPSECRET
#     from yao.method import write_file
#
#     amp = AMP(PROGRAMAPPID, PROGRAMAPPSECRET)
#     # token = amp.get_wxa_code(page="pages/news/auth", scene="3abafb7d21e747c192619566340f7509", state="trial")
#     # print(token)
#     # write_file("./aa.png", content=token)
#
#     # res = amp.message(openid="ozw6S67o8VFMSvgOjBDt3G6fMR9E", template_id="QZ9k9_RV5JlKEwUG5rmAH--AOVM5f8U0LKaxYHTim7k", data={
#     #     "thing2": {
#     #         "value": "页面打不开"
#     #     },
#     #     "character_string8": {
#     #         "value": "http://www.abc.com"
#     #     },
#     #     "thing10": {
#     #         "value": "就是打不开"
#     #     },
#     #     "time11": {
#     #         "value": "2019年10月1日 15:02"
#     #     }
#     # })
#
#     res = amp.custom_message(openid="ozw6S67o8VFMSvgOjBDt3G6fMR9E", content="QZ9k9_RV5JlKEwUG5rmAH--AOVM5f8U0LKaxYHTim7k")
#     print(res)



import importlib
from datetime import datetime, timedelta

from config import DEFAULT_FUNCTION_COMPANY, OAUTH_ADMIN_USERS
from yao.function.user.crud import CrudFunctionUser

from yao.function.user.schema import SchemasFunctionScopes

from yao.function.queue.type import queue_function_data
from yao.function.queue.crud import Crud as QueueCrud
from yao.function.queue.schema import SchemasStoreUpdate as QueueSchemasStoreUpdate, SchemasQueueAuth
from yao.db import session as _session


def add_queue_function(module, name, data=None, auth: SchemasFunctionScopes = None, priority=5, unique_key=None, scope=None) -> bool:
    """
    添加队列数据
    :param module:
    :param name:
    :param data:
    :param auth:
    :param priority:
    :param unique_key:
    :param scope:
    :return:
    """
    try:
        session = next(_session())
        if not auth:
            prefix_name = DEFAULT_FUNCTION_COMPANY.get("prefix_name", "yao")
            user_name = "%s@%s" % (prefix_name, OAUTH_ADMIN_USERS.keys()[0] if len(OAUTH_ADMIN_USERS) > 1 else "admin")
            user = CrudFunctionUser.init().first(session=session, where=[("username", user_name)])
            auth = SchemasFunctionScopes(user=user, prefix=user.prefix)
        data = queue_function_data(module, name, data, auth)
        item = QueueSchemasStoreUpdate(
            prefix=auth.prefix, username=auth.user.username, priority=priority,
            scope=scope, data=data, key=unique_key, queue_status=0, retry=0
        )
        if unique_key:
            QueueCrud.init().find_or_store_model(session=session, where=[("key", unique_key), ("queue_status", "in", [0, 3])], item=item, close=True)
        else:
            QueueCrud.init().store(session=session, item=item, close=True)
        return True
    except:
        return False


# def digestion_queue(where=[]):
#     """
#     消化队列
#     :return:
#     """
#     session = next(_session())
#     created_at = datetime.now() + timedelta(days=-7)
#     queues = QueueCrud.init().get(session=session, where=[("queue_status", "in", [0, 2]), ("retry", "<", 5), ("created_at", ">", created_at)] + where, order=[("priority", "asc")])
#     for queue in queues:
#         data = queue.data or {}
#         module_name = data.get("module", None)
#         metaclass = importlib.import_module(module_name)
#         if metaclass:
#             if data.get("type") == "function":
#                 function_name = data.get("function_name")
#                 function_data = data.get("function_data", {})
#                 fuc = getattr(metaclass, function_name)
#                 start_at = datetime.now()
#                 QueueCrud.init().update(
#                     session=session, uuid=queue.uuid, event=True, exclude_unset=True,
#                     item=QueueSchemasStoreUpdate(start_at=start_at, stop_at=None, queue_status=3, progress=None, progress_text=None)
#                 )
#                 try:
#                     auth = function_data.get("auth", {})
#                     auth_item = SchemasQueueAuth(**auth)
#                     if len(auth) > 0:
#                         del function_data["auth"]
#                     fuc(queue=queue.uuid, auth=auth_item, **function_data)
#                     queue_status = 1
#                     progress = "100%"
#                 except:
#                     queue_status = 2
#                     progress = None
#                 item = QueueSchemasStoreUpdate(queue_status=queue_status, progress=progress, stop_at=datetime.now(), retry=queue.retry + 1)
#                 QueueCrud.init().update(session=session, uuid=queue.uuid, item=item, event=True, exclude_unset=True)
#     session.close()


def digestion_queue(where=[]):
    """
    消化队列
    :return:
    """
    while True:
        session = next(_session())
        created_at = datetime.now() + timedelta(days=-7)
        queue = QueueCrud.init().first(
            session=session,
            where=[("queue_status", "in", [0, 2]), ("retry", "<", 5), ("created_at", ">", created_at)] + where,
            order=[("priority", "asc")]
        )
        if not queue:
            break
        else:
            data = queue.data or {}
            module_name = data.get("module", None)
            metaclass = importlib.import_module(module_name)
            if metaclass:
                if data.get("type") == "function":
                    function_name = data.get("function_name")
                    function_data = data.get("function_data", {})
                    fuc = getattr(metaclass, function_name)
                    start_at = datetime.now()
                    QueueCrud.init().update(
                        session=session, uuid=queue.uuid, event=True, exclude_unset=True, close=False,
                        item=QueueSchemasStoreUpdate(start_at=start_at, stop_at=None, queue_status=3, progress=None, progress_text=None)
                    )
                    try:
                        auth = function_data.get("auth", {})
                        auth_item = SchemasQueueAuth(**auth)
                        if len(auth) > 0:
                            del function_data["auth"]
                        fuc(queue=queue.uuid, auth=auth_item, **function_data)
                        queue_status = 1
                        progress = "100%"
                    except:
                        queue_status = 2
                        progress = None
                    item = QueueSchemasStoreUpdate(queue_status=queue_status, progress=progress, stop_at=datetime.now(), retry=queue.retry + 1)
                    QueueCrud.init().update(session=session, uuid=queue.uuid, item=item, event=True, close=False, exclude_unset=True)
        try:
            session.close()
        except:
            pass


def progress_queue(uuid, progress=None, progress_text=None):
    """
    更新进度
    :param uuid:
    :param progress:
    :param progress_text:
    :return:
    """
    try:
        if uuid:
            session = next(_session())
            QueueCrud.init().update(session=session, uuid=uuid, item=QueueSchemasStoreUpdate(progress=progress, progress_text=progress_text), event=True, exclude_unset=True)
            session.close()
    except:
        pass


# if __name__ == '__main__':
#     digestion_queue()



from fastapi import APIRouter

from yao.function.company.main import router as company
from yao.function.user.main import router as user
from yao.function.permission.main import router as permission
from yao.function.appointment.main import router as appointment
from yao.function.department.main import router as department
from yao.function.annex.main import router as annex
from yao.function.log.main import router as log
from yao.function.queue.main import router as queue

router = APIRouter()
router.include_router(company)
router.include_router(user)
router.include_router(permission)
router.include_router(appointment)
router.include_router(department)
router.include_router(annex)
router.include_router(log)
router.include_router(queue)



def token_verify_password(plain_password: str, hashed_password: str):
    """
    验证 oauth token密码
    :param plain_password: 明文密码
    :param hashed_password: hash 密码
    :return: bool
    """
    from passlib.context import CryptContext
    pwd_context = CryptContext(schemes=['bcrypt'], deprecated='auto')

    return False if not hashed_password else pwd_context.verify(plain_password, hashed_password)


def token_get_password_hash(password: str):
    """
    给 oauth user 加密
    :param password: 加密密码
    :return: hash
    """
    from passlib.context import CryptContext
    pwd_context = CryptContext(schemes=['bcrypt'], deprecated='auto')
    return pwd_context.hash(password)


def token_access_token(data: dict, key: str, algorithm: str, expires_delta):
    """
    生成 token
    :param data: 加密数据
    :param key: 加密key
    :param algorithm: 加密 算法
    :param expires_delta: 有效期 类型 timedelta
    :return: token
    """
    from datetime import datetime, timedelta
    from jose import jwt
    to_encode = data.copy()
    expire = datetime.now() + (expires_delta if expires_delta else timedelta(minutes=15))
    to_encode.update({"exp": expire})
    encoded_jwt = jwt.encode(claims=to_encode, key=key, algorithm=algorithm)
    return encoded_jwt


def token_payload(security_scopes, token, key, algorithm):
    """
    获取 当前授权用户数据
    :param security_scopes: SecurityScopes
    :param token: OAuth2PasswordBearer
    :param key: 加密key
    :param algorithm: 加密 算法
    :return:
    """
    from jose import jwt, JWTError
    from fastapi import HTTPException, status
    from pydantic import ValidationError
    if security_scopes.scopes:
        authenticate_value = f'Bearer scope="{security_scopes.scope_str}"'
    else:
        authenticate_value = f'Bearer'
    credentials_exception = HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                                          detail='Could not validate credentials',
                                          headers={"WWW-Authenticate": authenticate_value})
    try:
        payload = jwt.decode(token=token, key=key, algorithms=[algorithm])
        sub: str = payload.get('sub')
        if sub is None:
            raise credentials_exception
        token_scopes = payload.get("scopes", [])
    except (JWTError, ValidationError):
        raise credentials_exception
    """排除不授权限管理"""
    # Todo
    """排除不授权限管理"""
    for scope in security_scopes.scopes:
        if scope not in token_scopes:
            raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Not enough permissions",
                                headers={"WWW-Authenticate": authenticate_value})
    return payload


def store_permissions(permissions: list):
    """
    创建权限
    :param permissions:
    :return:
    """
    from yao.function.model import SYSTEM_PERMISSIONS
    from yao.db import session
    _session = next(session())
    # 检查控制台
    checkPermissionOrStore([{"name": "控制台", "scope": "Dashboard", "icon": "DataBoard"}], session=_session)
    # 检查其它权限
    checkPermissionOrStore(permissions, session=_session)
    # 检查系统权限
    checkPermissionOrStore(SYSTEM_PERMISSIONS, session=_session)


def checkPermissionOrStore(permissions: list, session=None, parent_pk: str = None):
    """
    检查或者创建权限
    :param permissions:
    :param session:
    :param parent_pk:
    :return:
    """
    from yao.function.permission.crud import CrudFunctionPermission
    from yao.function.permission.schema import SchemasFunctionStoreUpdate as SchemasFunctionPermissionStoreUpdate
    ACTION_ITEMS = [
        {"name": "列表", "scope": "list"},
        {"name": "详情", "scope": "get"},
        {"name": "创建", "scope": "store"},
        {"name": "更新", "scope": "update"},
        {"name": "删除", "scope": "delete"},
        {"name": "导出", "scope": "export"},
        {"name": "导入", "scope": "import"}
    ]

    permissionObj = CrudFunctionPermission.init()
    for permission in permissions:
        _scope = permission.get('scope')
        _name = permission.get('name')
        _action = permission.get('action', [])
        _children = permission.get('children', None)
        _is_menu = permission.get('is_menu', True)
        _is_action = permission.get('is_action', False)
        icon = permission.get('icon', None)
        parent = permissionObj.find_or_store_model(
            session=session,
            where=('scope', _scope), item=SchemasFunctionPermissionStoreUpdate(
                name=_name, scope=_scope, parent_id=parent_pk, is_menu=_is_menu, is_action=_is_action, icon=icon
            ), commit=True, refresh=False
        )
        if _children:
            checkPermissionOrStore(_children, session=session, parent_pk=parent.uuid)
        else:
            for action in (ACTION_ITEMS + _action):
                scope = "%s.%s" % (_scope, action.get('scope'))
                name = "%s %s" % (_name, action.get('name'))
                is_menu = action.get('is_menu', False)
                is_action = action.get('is_action', True)
                permissionObj.find_or_store_model(
                    session=session, where=('scope', scope), item=SchemasFunctionPermissionStoreUpdate(
                        name=name, scope=scope, parent_id=parent.uuid, is_menu=is_menu, is_action=is_action
                    ), commit=True, refresh=False
                )


def init_user_and_password(users: dict, company: dict):
    """
    重置账号密码
    :param users: {username:password}
    :param company: {"name": "site","prefix_name": "yao"}
    :return:
    """
    from yao.function.user.crud import CrudFunctionUser
    from yao.function.user.schema import SchemasFunctionUserStoreUpdate
    from yao.db import session
    _session = next(session())

    return [CrudFunctionUser.init().update_or_store_model(
        session=_session, where=('username', "%s@%s" % (company.get("prefix_name"), username)),
        item=SchemasFunctionUserStoreUpdate(username="%s@%s" % (company.get("prefix_name"), username), password=password, prefix=company.get("prefix_name")))
        for username, password in users.items()]


def init_function_company(company: dict):
    """
    重置账号密码
    :param company: {"name": "site","prefix_name": "yao"}
    :return:
    """
    from yao.function.company.crud import CrudFunctionCompany
    from yao.function.company.schema import SchemasFunctionStoreUpdate
    from yao.db import session
    _session = next(session())
    CrudFunctionCompany.init(cache=False).update_or_store_model(session=_session, where=('name', company.get("name")), item=SchemasFunctionStoreUpdate(**company))



from datetime import datetime
from typing import Optional, Union, List

from pydantic import BaseModel, validator

from config import SCHEMAS_SUCCESS_CODE, SCHEMAS_SUCCESS_STATUS, SCHEMAS_SUCCESS_MESSAGE, SCHEMAS_ERROR_CODE, SCHEMAS_ERROR_STATUS, SCHEMAS_ERROR_MESSAGE


class Schemas(BaseModel):
    """状态返回"""
    code: Optional[int] = SCHEMAS_SUCCESS_CODE
    status: Optional[str] = SCHEMAS_SUCCESS_STATUS
    message: Optional[str] = SCHEMAS_SUCCESS_MESSAGE
    data: Optional[Union[BaseModel, dict, list, str, bool, None]] = None


class SchemasError(BaseModel):
    """状态返回"""
    code: Optional[int] = SCHEMAS_ERROR_CODE
    status: Optional[str] = SCHEMAS_ERROR_STATUS
    message: Optional[str] = SCHEMAS_ERROR_MESSAGE
    data: Optional[Union[BaseModel, dict, list, str, bool, None]] = None


class SchemasPaginate(BaseModel):
    """分页"""
    items: Optional[list] = None  # 当前页的数据列表
    page: Optional[int] = None  # 当前页数
    pages: Optional[int] = None  # 总页数
    total: Optional[int] = None  # 总条数
    limit: Optional[int] = None  # 页条数


class ModelScreenParams(BaseModel):
    """获取列表默认参数"""
    page: Optional[int] = 1
    limit: Optional[int] = 25
    where: Optional[Union[dict, list]] = []
    join: Optional[Union[dict, list]] = []
    order: Optional[list] = []


class ModelScreenParamsForAll(BaseModel):
    """获取列表默认参数 不带分页"""
    where: Optional[Union[dict, list]] = []
    join: Optional[Union[dict, list]] = []
    order: Optional[list] = []


class SchemaPrefix(BaseModel):
    owns: Optional[str] = None

    @validator('owns')
    def p_owns(cls, name: str):
        return name.split("@", 1)[1] if name and len(name.split("@", 1)) == 2 else name


class SchemaPrefixNames(SchemaPrefix):
    name: Optional[str] = None

    @validator('name')
    def p_name(cls, name: str):
        return name.split("@", 1)[1] if name and len(name.split("@", 1)) == 2 else name


class SchemaParamsApi(SchemaPrefixNames):
    uuid: Optional[str] = None

    class Config:
        orm_mode = True


class ModelUUIDS(BaseModel):
    uuids: List[str] = None


class SchemasPrefixOwns(BaseModel):
    prefix: Optional[str] = None  # 公司前缀
    owns: Optional[str] = None  # 拥有


class SchemasAt(BaseModel):
    created_at: Optional[datetime] = None  # 创建时间
    updated_at: Optional[datetime] = None  # 更新时间
    deleted_at: Optional[datetime] = None  # 删除时间



from setuptools import setup, find_packages

with open("readme.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="yao",
    version="0.1.3",
    description="Dev",
    python_requires=">=3.8",
    author="Lsshu",
    author_email="admin@lsshu.cn",
    url="https://github.com/lsshu/yao",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPLv3",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
    ],
    install_requires=[
        "fastapi[all]==0.83.0",
        "python-jose[cryptography]==3.3.0",
        "passlib[bcrypt]",
        "SQLAlchemy==1.4.46",
        "Pillow",
        "PyMySQL==1.0.2",
        "sqlalchemy-mptt==0.2.5",
        "user-agents==2.2.0",
        "requests",
        "openpyxl",
        "aioredis",
        "filetype==1.2.0",
        "pycryptodome",
    ]
)



from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel

from yao.schema import SchemasPaginate, Schemas


class SchemasFunctionResponse(BaseModel):
    """权限 返回"""
    uuid: Optional[str] = None
    name: Optional[str] = None
    icon: Optional[str] = None
    scope: Optional[str] = None
    path: Optional[str] = None
    parent_id: Optional[str] = None
    is_menu: Optional[bool] = True
    is_action: Optional[bool] = True
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    children: Optional[List['SchemasFunctionResponse']] = None

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态

    class Config:
        orm_mode = True


class SchemasFunctionPaginateItem(SchemasPaginate):
    items: List[SchemasFunctionResponse]


class SchemasFunctionStoreUpdate(BaseModel):
    """授权角色 提交"""
    name: Optional[str] = None
    icon: Optional[str] = None
    scope: Optional[str] = None
    path: Optional[str] = None
    parent_id: Optional[str] = None
    is_menu: Optional[bool] = True
    is_action: Optional[bool] = True

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态


class SchemasFunctionMiniResponse(BaseModel):
    """权限 返回"""
    uuid: Optional[str] = None
    name: Optional[str] = None

    class Config:
        orm_mode = True


class SchemasFunctionMenuMiniResponse(SchemasFunctionMiniResponse):
    """权限 返回"""
    children: Optional[List['SchemasFunctionMenuMiniResponse']] = None

    class Config:
        orm_mode = True


class SchemasFunctionMenuMiniItemResponse(Schemas):
    """菜单树返回"""
    data: List[SchemasFunctionMenuMiniResponse]


class SchemasFunctionMenuResponse(BaseModel):
    """权限 返回"""
    uuid: Optional[str] = None
    name: Optional[str] = None
    icon: Optional[str] = None
    scope: Optional[str] = None
    path: Optional[str] = None
    children: Optional[List['SchemasFunctionMenuResponse']] = None

    class Config:
        orm_mode = True


class SchemasFunctionMenuStatusResponse(Schemas):
    """菜单树返回"""
    data: List[SchemasFunctionMenuResponse]



from datetime import datetime
from typing import Optional, List

from pydantic import BaseModel

from yao.schema import SchemasPaginate


class SchemasFunctionResponse(BaseModel):
    """模型 返回"""
    uuid: Optional[str] = None
    name: Optional[str] = None  # 公司名称
    prefix_name: Optional[str] = None  # 公司前缀
    contact_name: Optional[str] = None  # 联系人
    contact_phone: Optional[str] = None  # 联系电话

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True


class SchemasFunctionStoreUpdate(BaseModel):
    """模型 提交"""
    name: Optional[str] = None  # 公司名称
    prefix_name: Optional[str] = None  # 公司前缀
    contact_name: Optional[str] = None  # 联系人
    contact_phone: Optional[str] = None  # 联系电话

    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态


class SchemasFunctionPaginateItem(SchemasPaginate):
    items: List[SchemasFunctionResponse]


class SchemasParams(BaseModel):
    pass



from datetime import datetime
from typing import Optional, List
from pydantic import BaseModel, validator

from yao.schema import SchemaPrefixNames, SchemasPaginate


class SchemasFunctionResponse(SchemaPrefixNames):
    """角色 返回"""
    uuid: Optional[str] = None
    parent_uuid: Optional[str] = None
    name: Optional[str] = None
    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    class Config:
        orm_mode = True


class SchemasFunctionTreeResponse(SchemasFunctionResponse):
    children: Optional[List['SchemasFunctionTreeResponse']] = None


class SchemasFunctionPaginateItem(SchemasPaginate):
    items: Optional[List[SchemasFunctionTreeResponse]] = None


class SchemasFunctionStoreUpdate(BaseModel):
    """授权角色 提交"""
    prefix: Optional[str] = None
    parent_id: Optional[str] = None
    name: Optional[str] = None
    remarks: Optional[str] = None  # 备注
    sort: Optional[int] = None  # 排序
    status: Optional[bool] = None  # 状态


class SchemasParams(BaseModel):
    departments: Optional[List[SchemasFunctionTreeResponse]] = None



from sqlalchemy.orm import Session

from yao.crud import Operation
from yao.helpers import token_get_password_hash
from yao.function.model import ModelFunctionUsers, ModelFunctionPermissions, ModelFunctionAppointments


class CrudFunctionUser(Operation):
    """用户表操作"""
    model_class = ModelFunctionUsers

    relationships = {
        "permissions": ModelFunctionPermissions,
        "appointments": ModelFunctionAppointments
    }

    def store(self, item=None, data: dict = None, commit: bool = True, refresh: bool = True, close: bool = False, **kwargs):
        if hasattr(item, "password") and item.password:
            item.password = token_get_password_hash(item.password)
        item.children_ids = [item.username]
        res = super().store(item=item, data=data, commit=commit, refresh=refresh, close=close, **kwargs)
        return res

    def update(self, where=None, item=None, data: dict = None, commit: bool = True, refresh: bool = True, close: bool = False, exclude_unset=True, event: bool = False, **kwargs):
        if hasattr(item, "password") and item.password:
            item.password = token_get_password_hash(item.password)
        else:
            if hasattr(item, "password"):
                delattr(item, "password")
        res = super().update(where=where, item=item, data=data, commit=commit, refresh=refresh, close=False, exclude_unset=exclude_unset, event=event, **kwargs)
        self.update_children_ids(close=False, session=kwargs.get("session"))
        return res

    def update_children_ids(self, **kwargs):
        users = self.get(**kwargs)
        for user in users:
            ids = []

            def _get_id(obj):
                _ids = [obj.get("node").username]
                children = obj.get("children", [])
                if len(children) > 0:
                    for child in children:
                        _ids += _get_id(child)
                return _ids

            for _obj in user.drilldown_tree():
                ids += _get_id(_obj)
            user.children_ids = ids
        session = kwargs.get("session")
        commit = kwargs.get("commit", True)
        close = kwargs.get("close", True)
        commit and session.commit()
        close and session.close()
        return True



from typing import List

from fastapi import APIRouter, Depends, Security
from sqlalchemy.orm import Session

from yao.db import session as _session
from yao.depends import model_screen_params, model_post_screen_params, auth_user, route, item_name_prefix
from yao.schema import ModelScreenParams, Schemas, SchemasError
from yao.function.model import function_department_name as name
from yao.function.department.crud import CrudFunctionDepartment
from yao.function.department.schema import SchemasFunctionPaginateItem, SchemasParams, SchemasFunctionResponse, SchemasFunctionStoreUpdate
from yao.function.user.schema import SchemasFunctionScopes

router = APIRouter(tags=[name.replace('.', ' ').title()])

role_scopes = [name, ]


@route('/_M_', module=name, router=router, methods=['get'])
@item_name_prefix
async def get_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_screen_params),
                     auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.list" % name])):
    """
    :param session:
    :param params:
    :param auth:
    :return:
    """
    db_model_list = CrudFunctionDepartment.init().paginate(session=session, where=[('prefix', auth.prefix)], screen_params=params)
    return Schemas(data=SchemasFunctionPaginateItem(**db_model_list))


@route('/_M_.post', module=name, router=router, methods=['post'])
@item_name_prefix
async def post_models(session: Session = Depends(_session), params: ModelScreenParams = Depends(model_post_screen_params),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.list" % name])):
    """
    :param session:
    :param params:
    :param auth:
    :return:
    """
    db_model_list = CrudFunctionDepartment.init().paginate(session=session, screen_params=params, tree=True)
    return Schemas(data=SchemasFunctionPaginateItem(**db_model_list))


@route('/_M_.params', module=name, router=router, methods=['get'])
async def params_models(session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.list" % name])):
    """
    :param session:
    :param auth:
    :return:
    """
    data = {
        "departments": CrudFunctionDepartment.init().get_tree(session=session, where=[('prefix', auth.prefix)])
    }
    return Schemas(data=SchemasParams(**data))


@route('/_M_', module=name, router=router, methods=['post'])
@item_name_prefix
async def store_model(item: SchemasFunctionStoreUpdate, session: Session = Depends(_session),
                      auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.store" % name])):
    """
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionDepartment.init().first(session=session, where=("name", item.name))
    if db_model is not None:
        return SchemasError(message="数据已经存在！")
    if item.parent_id:
        parent_model = CrudFunctionDepartment.init().first(session=session, uuid=item.parent_id)
        if parent_model:
            item.parent_id = parent_model.id
    bool_model = CrudFunctionDepartment.init().store(session=session, item=item)
    return Schemas(data=SchemasFunctionResponse(**bool_model.to_dict()))


@route("/_M_/{uuid}", module=name, router=router, methods=['patch'])
@item_name_prefix
async def update_patch_model(uuid: str, item: SchemasFunctionStoreUpdate, session: Session = Depends(_session),
                             auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.update" % name])):
    """
    :param uuid:
    :param item:
    :param session:
    :param auth:
    :return:
    """
    db_model = CrudFunctionDepartment.init().first(session=session, uuid=uuid)
    if db_model is None:
        return SchemasError(message="数据没有找到！")
    CrudFunctionDepartment.init().update(session=session, uuid=uuid, item=item, exclude_unset=True)
    return Schemas()


@route("/_M_", module=name, router=router, methods=['delete'])
async def delete_models(uuids: List[str], session: Session = Depends(_session), auth: SchemasFunctionScopes = Security(auth_user, scopes=role_scopes + ["%s.delete" % name])):
    """
    :param uuids:
    :param session:
    :param auth:
    :return:
    """
    bool_model = CrudFunctionDepartment.init().delete(session=session, uuids=uuids, event=True)
    return Schemas(data=bool_model)
